"""Custom optimizers."""

from torch import optim

BETA1 = 0.9
BETA2 = 0.999


# The Adam family of optimizers use a pair of numerical coefficients in the
# range [0, 1), usually denoted by \beta or (\beta_1, \beta_2), to keep a
# running average of the gradients. Idiotically, the PyTorch implementation
# passes these hyperparameters as a tuple of two floats rather than as two
# separate floats. While this is already annoying, it is mainly a problem if
# one wishes to tune these two values using the W&B hyperparameter sampling
# system, because this engine samples scalar values, not tuples. Any way to
# convert the scalar to a tuple is necessarily going to be a hack.
#
# Here we minimally modify the optimizer constructors so that one passes
# `beta1: float` and `beta2: float` in place of `betas: Tuple[float, float]`.
# This is implemented by subclassing each one separately rather than
# fancy metaprogramming. The one complexity is that it is necesssary to
# explicitly copy the signature of the function up to but not including
# `betas`, which includes the parameter tensor and the default learning rate,
# which varies from optimizer to optimizer; the rest can be done safely with
# `*args` and `**kwargs.
#
# One can use these subclasses if separate betas are desired. For example
# use, see examples/wandb_sweeps.


class Adam(optim.Adam):
    """Adam optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.001,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)


class AdamW(optim.AdamW):
    """AdamW optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.001,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)


class Adamax(optim.Adamax):
    """Adamax optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.002,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)


class NAdam(optim.NAdam):
    """Adamax optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.002,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)


class RAdam(optim.RAdam):
    """RAdam optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.001,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)


class SparseAdam(optim.SparseAdam):
    """Sparse Adam optimizer with separate betas."""

    def __init__(
        self,
        params,
        lr: float = 0.001,
        beta1: float = BETA1,
        beta2: float = BETA2,
        *args,
        **kwargs,
    ):
        super().__init__(params, lr, (beta1, beta2), *args, **kwargs)



"""Pointer-generator model classes."""

from typing import Callable, Optional, Tuple, Union

import torch
from torch import nn

from .. import data, special
from . import base, beam_search, defaults, modules, rnn, transformer


class PointerGeneratorModel(base.BaseModel):
    """Base class for pointer-generator models."""

    # Constructed inside __init__.
    geneneration_probability: modules.GenerationProbability

    def _get_loss_func(
        self,
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Returns the actual function used to compute loss.

        This overrides the loss function behavior in
        models.base.BaseModel because we need to use NLLLoss in
        order to factor the addition of two separate probability
        distributions. An NLLLoss-compatible implementation of label smoothing
        is also provided when label smoothing is requested.

        Returns:
            Callable[[torch.Tensor, torch.Tensor], torch.Tensor]: configured
                loss function.
        """
        if not self.label_smoothing:
            return nn.NLLLoss(ignore_index=special.PAD_IDX)
        else:
            return self._smooth_nllloss

    def _smooth_nllloss(
        self, predictions: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Computes the NLLLoss with a smoothing factor such that some
        proportion of the output distribution is replaced with a
        uniform distribution.

        After:

            https://github.com/NVIDIA/DeepLearningExamples/blob/
            8d8b21a933fff3defb692e0527fca15532da5dc6/PyTorch/Classification/
            ConvNets/image_classification/smoothing.py#L18

        Args:
            predictions (torch.Tensor): tensor of prediction
                distribution of shape B x target_vocab_size x seq_len.
            target (torch.Tensor): tensor of golds of shape
                B x seq_len.

        Returns:
            torch.Tensor: loss.
        """
        # -> (B * seq_len) x target_vocab_size.
        predictions = predictions.transpose(1, 2).reshape(
            -1, self.target_vocab_size
        )
        # -> (B * seq_len) x 1.
        target = target.view(-1, 1)
        non_pad_mask = target.ne(special.PAD_IDX)
        # Gets the ordinary loss.
        nll_loss = -predictions.gather(dim=-1, index=target)[
            non_pad_mask
        ].mean()
        # Gets the smoothed loss.
        smooth_loss = -predictions.sum(dim=-1, keepdim=True)[
            non_pad_mask
        ].mean()
        smooth_loss = smooth_loss / self.target_vocab_size
        # Combines both according to label smoothing weight.
        loss = (1.0 - self.label_smoothing) * nll_loss
        loss.add_(self.label_smoothing * smooth_loss)
        return loss


class PointerGeneratorRNNModel(PointerGeneratorModel, rnn.RNNModel):
    """Abstract base class for pointer-generator models with RNN backends.

    If features are provided, a separate features attention module computes
    the feature encodings which are then concatenated with the source attention
    output on the encoding dimension.

    After:
        See, A., Liu, P. J., and Manning, C. D. 2017. Get to the point:
        summarization with pointer-generator networks. In Proceedings of the
        55th Annual Meeting of the Association for Computational Linguistics
        (Volume 1: Long Papers), pages 1073-1083.

    Args:
        *args: passed to superclass.
        **kwargs: passed to superclass.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Uses the inherited defaults for the source embeddings and encoder.
        if self.has_features_encoder:
            self.features_attention = modules.Attention(
                self.features_encoder.output_size, self.decoder_hidden_size
            )
        self.generation_probability = modules.GenerationProbability(
            self.embedding_size,
            self.decoder_hidden_size,
            self.decoder_input_size,
        )
        self.classifier = nn.Linear(
            self.decoder_hidden_size + self.decoder_input_size,
            self.target_vocab_size,
        )
        # Compatibility check.
        if self.source_encoder.layers != self.decoder_layers:
            raise base.ConfigurationError(
                f"Number of encoder layers ({self.source_encoder.layers}) and "
                f"decoder layers ({self.decoder_layers}) must match"
            )

    def beam_decode(
        self,
        source: torch.Tensor,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        features_encoded: Optional[torch.Tensor] = None,
        features_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Decodes with beam search.

        Implementationally this is almost identical to the method of the same
        name in RNNModel.

        Args:
            source (torch.Tensor): source symbols, used to compute pointer
                weights.
            source_encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for the source.
            features_encoded (torch.Tensor, optional): encoded feaure symbols.
            features_mask (torch.Tensor, optional): mask for the features.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: predictions of shape
                B x beam_width x seq_length and log-likelihoods of shape
                B x beam_width.
        """
        # TODO: modify to work with batches larger than 1.
        batch_size = source_mask.size(0)
        if batch_size != 1:
            raise NotImplementedError(
                "Beam search is not implemented for batch_size > 1"
            )
        state = self.decoder.initial_state(batch_size)
        # The start symbol is not needed here because the beam puts that in
        # automatically.
        beam = beam_search.Beam(self.beam_width, state)
        for _ in range(self.max_target_length):
            for cell in beam.cells:
                if cell.final:
                    beam.push(cell)
                else:
                    symbol = torch.tensor([[cell.symbol]], device=self.device)
                    prediction, state = self.decode_step(
                        source,
                        source_encoded,
                        source_mask,
                        symbol,
                        cell.state,
                        features_encoded,
                        features_mask,
                    )
                    scores = nn.functional.log_softmax(
                        prediction.squeeze(), dim=0
                    )
                    for new_cell in cell.extensions(state, scores):
                        beam.push(new_cell)
            beam.update()
            if beam.final:
                break
        return beam.predictions(self.device), beam.scores(self.device)

    def decode_step(
        self,
        source: torch.Tensor,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        symbol: torch.Tensor,
        state: modules.RNNState,
        features_encoded: Optional[torch.Tensor] = None,
        features_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Single decoder step.

        This predicts a distribution for one symbol.

        Args:
            source (torch.Tensor): source symbols, used to compute pointer
                weights.
            source_encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for the source.
            symbol (torch.Tensor): next symbol.
            state (modules.RNNState): RNN state.
            features_encoded (torch.Tensor, optional): encoded features
                symbols.
            features_mask (torch.Tensor, optional): mask for the features.

        Returns:
            Tuple[torch.Tensor, modules.RNNState]: predictions for that state
                and the RNN state.
        """
        # TODO: there are a few Law of Demeter violations here. Is there an
        # obvious refactoring?
        embedded = self.decoder.embed(symbol, self.embeddings)
        context, attention_weights = self.decoder.attention(
            source_encoded,
            state.hidden.transpose(0, 1),
            source_mask,
        )
        if self.has_features_encoder:
            features_context, _ = self.features_attention(
                features_encoded,
                state.hidden.transpose(0, 1),
                features_mask,
            )
            # -> B x 1 x 4*hidden_size.
            context = torch.cat((context, features_context), dim=2)
        _, state = self.decoder.module(
            torch.cat((embedded, context), dim=2), state
        )
        # -> B x 1 x hidden_size.
        hidden = state.hidden[-1, :, :].unsqueeze(1)
        output_dist = nn.functional.softmax(
            self.classifier(torch.cat((hidden, context), dim=2)),
            dim=2,
        )
        # -> B x 1 x target_vocab_size.
        pointer_dist = torch.zeros(
            (symbol.size(0), 1, self.target_vocab_size),
            device=self.device,
            dtype=attention_weights.dtype,
        )
        # Gets the attentions to the source in terms of the output generations.
        # These are the "pointer" distribution.
        pointer_dist.scatter_add_(2, source.unsqueeze(1), attention_weights)
        # Probability of generating from output_dist.
        gen_probs = self.generation_probability(context, hidden, embedded)
        scaled_output_dist = output_dist * gen_probs
        scaled_pointer_dist = pointer_dist * (1 - gen_probs)
        return torch.log(scaled_output_dist + scaled_pointer_dist), state

    def forward(
        self,
        batch: data.Batch,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]: beam search
                returns a tuple with a tensor of predictions of shape
                B x beam_width x seq_len and a tensor of B x shape beam_width
                with the likelihood (the unnormalized sum of sequence
                log-probabilities) for each prediction; greedy search returns
                a tensor of predictions of shape
                B x target_vocab_size x seq_len.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Features column specified but no feature
                encoder specified.
        """
        source_encoded = self.source_encoder(batch.source, self.embeddings)
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.Error(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            if self.beam_width > 1:
                return self.beam_decode(
                    batch.source.padded,
                    source_encoded,
                    batch.source.mask,
                    features_encoded=features_encoded,
                    features_mask=batch.features.mask,
                )
            else:
                return self.greedy_decode(
                    batch.source.padded,
                    source_encoded,
                    batch.source.mask,
                    teacher_forcing=(
                        self.teacher_forcing if self.training else False
                    ),
                    target=batch.target.padded if batch.has_target else None,
                    features_encoded=features_encoded,
                    features_mask=batch.features.mask,
                )
        elif batch.has_features:
            raise base.Error(
                "Features column specified but no feature encoder specified"
            )
        elif self.beam_width > 1:
            return self.beam_decode(
                batch.source.padded, source_encoded, batch.source.mask
            )
        else:
            return self.greedy_decode(
                batch.source.padded,
                source_encoded,
                batch.source.mask,
                teacher_forcing=(
                    self.teacher_forcing if self.training else False
                ),
                target=batch.target.padded if batch.has_target else None,
            )

    def greedy_decode(
        self,
        source: torch.Tensor,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        teacher_forcing: bool = defaults.TEACHER_FORCING,
        target: Optional[torch.Tensor] = None,
        features_encoded: Optional[torch.Tensor] = None,
        features_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Decodes a sequence given the encoded input.

        A hard upper bound on the length of the decoded strings is provided by
        the length of the target strings (more specifically, the length of the
        longest target string) if a target is provided, or `max_target_length`
        if not. Decoding will halt earlier if no target is provided and all
        sequences have reached END.

        Implementationally this is almost identical to the method of the same
        name in RNNModel.

        Args:
            source (torch.Tensor): source symbols, used to compute pointer
                weights.
            source_encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for the source.
            teacher_forcing (bool, optional): whether or not to decode with
                teacher forcing.
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues up until this length is reached.
            features_encoded (torch.Tensor, optional): encoded feaure symbols.
            features_mask (torch.Tensor, optional): mask for the features.

        Returns:
            torch.Tensor: predictions of B x target_vocab_size x seq_len.
        """
        batch_size = source_mask.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        predictions = []
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        for t in range(max_num_steps):
            prediction, state = self.decode_step(
                source,
                source_encoded,
                source_mask,
                symbol,
                state,
                features_encoded,
                features_mask,
            )
            predictions.append(prediction.squeeze(1))
            # With teacher forcing the next input is the gold symbol for this
            # step; with student forcing, it's the top prediction.
            symbol = (
                target[:, t].unsqueeze(1)
                if teacher_forcing
                else torch.argmax(prediction, dim=2)
            )
            if target is None:
                # Updates which sequences have decoded an END.
                final = torch.logical_or(final, symbol == special.END_IDX)
                if final.all():
                    break
        predictions = torch.stack(predictions, dim=2)
        return predictions

    @property
    def decoder_input_size(self) -> int:
        # We concatenate along the encoding dimension.
        if self.has_features_encoder:
            return (
                self.source_encoder.output_size
                + self.features_encoder.output_size
            )
        else:
            return self.source_encoder.output_size


class PointerGeneratorGRUModel(PointerGeneratorRNNModel):
    """Pointer-generator model with an GRU backend."""

    def get_decoder(self) -> modules.SoftAttentionGRUDecoder:
        return modules.SoftAttentionGRUDecoder(
            attention_input_size=self.source_encoder.output_size,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "pointer-generator GRU"


class PointerGeneratorLSTMModel(PointerGeneratorRNNModel):
    """Pointer-generator model with an LSTM backend."""

    def get_decoder(self) -> modules.SoftAttentionLSTMDecoder:
        return modules.SoftAttentionLSTMDecoder(
            attention_input_size=self.source_encoder.output_size,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "pointer-generator LSTM"


class PointerGeneratorTransformerModel(
    PointerGeneratorModel, transformer.TransformerModel
):
    """Pointer-generator model with a transformer backend.

    After:
        Singer, A., and Kann, K. 2020. The NYU-CUBoulder Systems for
        SIGMORPHON 2020 Task 0 and Task 2. In Proceedings of the 17th
        SIGMORPHON Workshop on Computational Research in Phonetics, Phonology,
        and Morphology, pages 90–98.

    Args:
        attention_heads (int).
        *args: passed to the superclass.
        **kwargs: passed to the superclass.
    """

    # Model arguments.
    attention_heads: int

    def __init__(
        self,
        *args,
        attention_heads=defaults.ATTENTION_HEADS,
        **kwargs,
    ):
        self.attention_heads = attention_heads
        super().__init__(
            *args,
            **kwargs,
        )
        if self.has_features_encoder:
            self.generation_probability = modules.GenerationProbability(
                self.embedding_size,
                self.embedding_size,
                self.embedding_size,
            )
        else:
            self.generation_probability = modules.GenerationProbability(
                self.embedding_size,
                self.embedding_size,
                self.source_encoder.output_size,
            )

    def decode_step(
        self,
        source: torch.Tensor,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        target: torch.Tensor,
        features_encoded: Optional[torch.Tensor] = None,
        features_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Single decoder step.

        This will work on any sequence length, and returns output
        probabilities for all targets, meaning we can use this method for
        greedy decoding, wherein only a single new token is decoded at a time,
        or for teacher-forced training, wherein all tokens can be decoded in
        parallel with a diagonal mask.

        Args:
            source (torch.Tensor): source symbols, used to compute pointer
                weights.
            source_encoded (torch.Tensor): encoded source_symbols.
            source_mask (torch.Tensor): mask for the source.
            target (torch.Tensor): tensor of predictions thus far.
            features_encoded (torch.Tensor, optional): encoded features
                symbols.
            features_mask (torch.Tensor, optional): mask for the features.

        Returns:
            torch.Tensor: predictions for that state.
        """
        # Uses a dummy mask of all zeros.
        target_mask = torch.zeros_like(target, dtype=bool)
        decoded, target_embedded = self.decoder(
            source_encoded,
            source_mask,
            target,
            target_mask,
            self.embeddings,
            features_encoded,
            features_mask,
        )
        # Outputs from the multi-headed attention from each decoder step to
        # the encoded source. Values have been averaged over each attention
        # head.
        # -> B x target_seq_len x source_seq_len.
        mha_outputs = self.decoder.attention_output[0]
        # Clears the stored attention result.
        self.decoder.attention_output.clear()
        logits = self.classifier(decoded)
        output_dist = nn.functional.softmax(logits, dim=2)
        # -> B x target-seq_len x target_vocab_size.
        pointer_dist = torch.zeros(
            mha_outputs.size(0),
            mha_outputs.size(1),
            self.target_vocab_size,
            device=self.device,
            dtype=mha_outputs.dtype,
        )
        # Repeats the source indices for each target.
        # -> B x target_seq_len x source_seq_len.
        repeated_source = source.unsqueeze(1).repeat(1, mha_outputs.size(1), 1)
        # Scatters the attention weights onto the pointer_dist at their vocab
        # indices in order to get outputs that match the indexing of the
        # generation probability.
        pointer_dist.scatter_add_(2, repeated_source, mha_outputs)
        # A matrix of context vectors from applying attention to the encoder
        # representations w.r.t. each decoder step.
        context = torch.bmm(mha_outputs, source_encoded)
        # Probability of generating from output_dist.
        gen_probs = self.generation_probability(
            context, decoded, target_embedded
        )
        scaled_pointer_dist = pointer_dist * (1 - gen_probs)
        scaled_output_dist = output_dist * gen_probs
        return torch.log(scaled_output_dist + scaled_pointer_dist)

    def forward(
        self,
        batch: data.Batch,
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            torch.Tensor.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Features column specified but no feature
                encoder specified.
            NotImplementedError: Beam search not implemented.
        """
        source_encoded = self.source_encoder(batch.source, self.embeddings)
        if self.has_features_encoder:
            if not batch.features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            return self.greedy_decode(
                batch.source.padded,
                source_encoded,
                batch.source.mask,
                batch.target.padded if batch.has_target else None,
                features_encoded=features_encoded,
                features_mask=batch.features.mask,
            )
        elif batch.has_features:
            raise base.ConfigurationError(
                "Feature column specified but no feature encoder specified"
            )
        else:
            return self.greedy_decode(
                batch.source.padded,
                source_encoded,
                batch.source.mask,
                batch.target.padded if batch.has_target else None,
            )

    def get_decoder(
        self,
    ) -> modules.TransformerPointerDecoder:
        return modules.TransformerPointerDecoder(
            attention_heads=self.attention_heads,
            decoder_input_size=self.source_encoder.output_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            has_features_encoder=self.has_features_encoder,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            max_length=self.decoder_max_length,
            num_embeddings=self.num_embeddings,
        )

    def greedy_decode(
        self,
        source: torch.Tensor,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        target: Optional[torch.Tensor] = None,
        features_encoded: Optional[torch.Tensor] = None,
        features_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Decodes the output sequence greedily.

        A hard upper bound on the length of the decoded strings is provided by
        the length of the target strings (more specifically, the length of the
        longest target string) if a target is provided, or `max_target_length`
        if not. Decoding will halt earlier if no target is provided and all
        sequences have reached END.

        Args:
            source (torch.Tensor): source symbols, used to compute pointer
                weights.
            source_encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for the source.
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues until this length is reached.
            features_encoded (torch.Tensor, optional): encoded feaure symbols.
            features_mask (torch.Tensor, optional): mask for the features.

        Returns:
            torch.Tensor: predictions of B x target_vocab_size x seq_len.
        """
        batch_size = source_encoded.size(0)
        # The output distributions to be returned.
        outputs = []
        # The predicted symbols at each iteration.
        predictions = [
            torch.tensor([special.START_IDX], device=self.device).repeat(
                batch_size
            )
        ]
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        for _ in range(max_num_steps):
            scores = self.decode_step(
                source,
                source_encoded,
                source_mask,
                torch.stack(predictions, dim=1),
                features_encoded,
                features_mask,
            )
            scores = scores[:, -1, :]
            outputs.append(scores)
            symbol = torch.argmax(scores, dim=1)
            predictions.append(symbol)
            if target is None:
                # Updates which sequences have decoded an END.
                final = torch.logical_or(final, (symbol == special.END_IDX))
                if final.all():
                    break
        outputs = torch.stack(outputs, dim=2)
        return outputs

    @property
    def name(self) -> str:
        return "pointer-generator transformer"



"""Finds the best batch size given constraints."""

import logging
import math
from typing import Tuple

import lightning
from lightning.pytorch.tuner import tuning

from . import data, defaults, models


class Error(Exception):
    pass


def _optimal_batch_size(
    desired_batch_size: int, max_batch_size: int
) -> Tuple[int, int]:
    r"""Computes optimal batch size and number of gradient accumulation steps.

    Given the desired batch size $b$ and a max batch size $n_{\textrm{max}}$,
    this returns a tuple $a \in \mathcal{N}_{+}, n \in
    \mathcal{N}_{\le n_\textrm{max}}$ such that $n$ is the largest integer
    $\le n_\textrm{max}$ and $a n = b$.

    If desired batch size` is smaller than max batch size, then we just use
    desired batch size without multiple steps of gradient accumulation.

    Args:
        desired_batch_size (int).
        max_batch_size (int).

    Returns:
        Tuple[int, int]: the tuple (a, n) as defined above.
    """
    # If the max size is larger than the non-zero desired size, use the
    # desired size.
    if desired_batch_size <= max_batch_size:
        return 1, desired_batch_size
    # This simulates the desired batch size with multiple steps of gradient
    # accumulation per update. The solution given here is a brute force one,
    # but pilot experiments with a more elegant solution using the divisors was
    # no more efficient.
    for batch_size in range(max_batch_size + 1, 0, -1):
        if desired_batch_size % batch_size == 0:
            accum_steps = desired_batch_size // batch_size
            return accum_steps, batch_size


def find_batch_size(
    method: str,
    trainer: lightning.Trainer,
    model: models.BaseModel,
    datamodule: data.DataModule,
    *,
    steps_per_trial: int = defaults.FIND_BATCH_SIZE_STEPS_PER_TRIAL,
) -> None:
    """Computes maximum or optimal batch size.

    This sets the batch size on the datamodule, and if necessary, the
    gradient accumulation steps on the trainer.

    With the max method, batch size is doubled until a memory error, thus
    finding the largest batch size that fits in memory.

    With the optimal method, batch size is doubled until a memory error or
    once we confirm that the desired batch size fits in memory (whichever
    comes first).

    Args:
        method (str): one of "max" (find the maximum batch size) or "opt" (find
            the "optimal" batch size, using the gradient accumulation trick if
            necessary.)
        trainer (lightning.Trainer).
        model (models.BaseModel).
        datamodule (data.DataModule).
        steps_per_trial (int, optional): number of steps to run with a given
            batch size.
    """
    tuner = tuning.Tuner(trainer)
    if method == "max":
        max_batch_size = tuner.scale_batch_size(
            model,
            datamodule=datamodule,
            steps_per_trial=steps_per_trial,
        )
        # This automatically overwrites the datamodule batch size argument.
        # So we don't need to.
    elif method == "opt":
        # This is a copy since it'll be automatically overwritten with the
        # maximum value.
        desired_batch_size = datamodule.batch_size
        if desired_batch_size <= 2:
            raise Error("--find_batch_size opt requires --batch_size > 2")
        max_batch_size = tuner.scale_batch_size(
            model,
            datamodule=datamodule,
            # This computes the maximum number of steps that'll be needed to
            # exceed the desired batch size. This seems like it would be "off
            # by one" but the way it's defined in the batch size finder is
            # itself "off by one" in the opposite direction, so it cancels out.
            max_trials=int(math.log2(desired_batch_size - 1)),
            steps_per_trial=steps_per_trial,
        )
        steps, batch_size = _optimal_batch_size(
            desired_batch_size, max_batch_size
        )
        datamodule.batch_size = batch_size
        logging.info("Using optimal batch size: %d", datamodule.batch_size)
        if steps != 1:
            trainer.accumulate_grad_batches = steps
            logging.info(
                "Using gradient accumulation steps: %d",
                trainer.accumulate_grad_batches,
            )
    else:
        raise Error(f"Unknown batch sizing method: {method}")



"""Utilities."""

import logging
import os
from typing import Any, Dict

import torch
import yaml

from . import special


def mkpath(path: str) -> None:
    """Creates subdirectories for a path if they do not already exist.

    Args:
        path (str).
    """
    dirname = os.path.dirname(os.path.abspath(path))
    os.makedirs(dirname, exist_ok=True)


def pad_tensor_after_end(
    predictions: torch.Tensor,
) -> torch.Tensor:
    """Replaces everything after an END token with PADs.

    Cuts off tensors at the first END, and replaces the rest of the
    predictions with PAD_IDX, as these can be erroneously decoded while the
    rest of the batch is finishing decoding.

    Args:
        predictions (torch.Tensor): prediction tensor.

    Returns:
        torch.Tensor: finalized predictions.
    """
    # Not necessary if batch size is 1.
    if predictions.size(0) == 1:
        return predictions
    for i, prediction in enumerate(predictions):
        # Gets first instance of END.
        end = (prediction == special.END_IDX).nonzero(as_tuple=False)
        if len(end) > 0 and end[0].item() < len(prediction):
            # If an END was decoded and it is not the last one in the
            # sequence.
            end = end[0]
        else:
            # Leaves predictions[i] alone.
            continue
        # Hack in case the first prediction is END. In this case
        # torch.split will result in an error, so we change these 0's to
        # 1's, which will make the entire sequence END as intended.
        end[end == 0] = 1
        symbols, *_ = torch.split(prediction, end)
        # Replaces everything after with PAD, to replace erroneous decoding
        # While waiting on the entire batch to finish.
        pads = (
            torch.ones(len(prediction) - len(symbols), device=symbols.device)
            * special.PAD_IDX
        )
        pads[0] = special.END_IDX
        # Makes an in-place update to an inference tensor.
        with torch.inference_mode():
            predictions[i] = torch.cat((symbols, pads))
    return predictions


def load_config(path: str) -> Dict[str, Any]:
    """Loads a YAML config file, parsing the LINKS field separately.

    Args:
        path (str).

    Returns:
        The config file as a dictionary.
    """
    with open(path, "r") as source:
        return yaml.safe_load(source)


def recursive_insert(config: Dict[str, Any], key: str, value) -> None:
    """Recursively inserts values into a nested dictionary.

    Args:
        config: the config dictionary.
        key: a string with the arguments separated by ".".
        value: the value to insert.
    """
    *most, last = key.split(".")
    ptr = config
    for piece in most:
        try:
            ptr = ptr[piece]
        except KeyError:
            ptr[piece] = {}
            ptr = ptr[piece]
    if last in ptr:
        logging.debug(
            "Overriding configuration argument %s with W&B sweep value: %r",
            key,
            value,
        )
    ptr[last] = value



"""Linear module class."""

from ... import data, defaults
from . import base

import torch
from torch import nn


class LinearEncoder(base.BaseModule):
    """Embeds the input tensor, and then applies a affine projection.

    This produces a simple non-contextual encoding of the input tensor.

    Args:
        *args: passed to superclass.
        output_size (int, optional).
        **kwargs: passed to superclass.
    """

    linear: nn.Linear

    def __init__(
        self, *args, output_size: int = defaults.HIDDEN_SIZE, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.linear = nn.Linear(self.embedding_size, output_size)

    def forward(
        self, symbols: data.PaddedTensor, embeddings: nn.Embedding
    ) -> torch.Tensor:
        """Encodes the input.

        Args:
            symbols (data.PaddedTensor): padded tensor.
            embeddings (nn.Embedding): embeddings.

        Returns:
            torch.Tensor.
        """
        return self.linear(self.dropout_layer(embeddings(symbols.padded)))

    @property
    def name(self) -> str:
        return "linear"

    @property
    def output_size(self) -> int:
        return self.linear.out_features



"""Base model class, with PL integration."""

import abc
import logging
from typing import Callable, Optional, Tuple, Union

import lightning
from lightning.pytorch import cli

import torch
from torch import nn, optim

from .. import data, defaults, metrics, special
from . import modules


class ConfigurationError(ValueError):
    pass


class BaseModel(abc.ABC, lightning.LightningModule):
    """Abstract base class for models handling Lightning integration.

    The following are defaults, but can be overriden by individual models:

    * The forward method returns a tensor of shape B x target_vocab_size x
      seq_length for compatibility with loss and evaluation functions unless
      beam search is enabled.
    * Cross-entropy loss is the loss function.
    * One or more predictions tensor(s) are returned by predict_step.
    * Loss is returned by training_step.
    * Evaluation metrics are tracked by test_step; nothing is returned.
    * Validation loss and evaluation metrics are tracked by validation_step;
      nothing is returned.
    * If features_encoder is True, the source encoder will be reused as the
      features encoder and if False (the default), no features encoder will be
      used.

    Unknown positional or keyword args from the superclass are ignored.

    Args:
        source_encoder (modules.BaseModule).
        decoder_hidden_size (int, optional): dimensionality of decoder layers.
        decoder_layers (int, optional): number of decoder layers.
        decoder_dropout (float, optional): dropout probability.
        embedding_size (int, optional): dimensionality of embedding.
        features_encoder (modules.BaseModule, optional).
        label_smoothing (float, optional): label smoothing coefficient.
        max_target_length (int, optional): maximum target length.
    """

    beam_width: int
    decoder_layers: int
    decoder_hidden_size: int
    decoder_dropout: float
    label_smoothing: float
    optimizer: optim.Optimizer
    scheduler: optim.lr_scheduler.LRScheduler
    source_encoder: modules.BaseModule
    features_encoder: Optional[modules.BaseModule]
    accuracy: Optional[metrics.Accuracy]
    ser: Optional[metrics.SER]
    decoder: modules.BaseModule
    embedding: nn.Embedding
    loss_func: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]

    def __init__(
        self,
        source_encoder: modules.BaseModule,
        *args,  # Ignored here.
        beam_width: int = defaults.BEAM_WIDTH,
        compute_accuracy: bool = True,
        compute_ser: bool = False,
        decoder_hidden_size: int = defaults.HIDDEN_SIZE,
        decoder_layers: int = defaults.LAYERS,
        decoder_dropout: float = defaults.DROPOUT,
        embedding_size: int = defaults.EMBEDDING_SIZE,
        features_encoder: Union[modules.BaseModule, bool] = False,
        label_smoothing: float = defaults.LABEL_SMOOTHING,
        max_target_length: int = defaults.MAX_LENGTH,
        optimizer: cli.OptimizerCallable = defaults.OPTIMIZER,
        scheduler: cli.LRSchedulerCallable = defaults.SCHEDULER,
        target_vocab_size: int = -1,  # Dummy value filled in via link.
        vocab_size: int = -1,  # Dummy value filled in via link.
        **kwargs,  # Ignored here.
    ):
        super().__init__()
        self.beam_width = beam_width
        self.decoder_hidden_size = decoder_hidden_size
        self.decoder_layers = decoder_layers
        self.decoder_dropout = decoder_dropout
        self.embedding_size = embedding_size
        self.label_smoothing = label_smoothing
        self.max_target_length = max_target_length
        self.num_embeddings = vocab_size
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.target_vocab_size = target_vocab_size
        self.accuracy = (
            metrics.Accuracy(self.target_vocab_size)
            if compute_accuracy
            else None
        )
        self.ser = metrics.SER() if compute_ser else None
        self.embeddings = self.init_embeddings(
            self.num_embeddings, self.embedding_size
        )
        if source_encoder.embedding_size != self.embedding_size:
            raise ConfigurationError(
                "Source embedding size "
                f"({source_encoder.embedding_size}) != "
                "model embedding size "
                f"({self.embedding_size})"
            )
        self.source_encoder = source_encoder
        if features_encoder is True:
            self.features_encoder = self.source_encoder
            self.has_features_encoder = True
        elif features_encoder is False:
            self.feature_encoder = None
            self.has_features_encoder = False
        else:
            if features_encoder.embedding_size != self.embedding_size:
                raise ConfigurationError(
                    "Features embedding size "
                    f"({features_encoder.embedding_size}) != "
                    "model embedding size "
                    f"({self.embedding_size})"
                )
            self.features_encoder = features_encoder
            self.has_features_encoder = True
        self.decoder = self.get_decoder()
        self.loss_func = self._get_loss_func()
        self.save_hyperparameters(
            ignore=[
                "source_encoder",
                "features_encoder",
                "decoder",
            ]
        )
        self._log_model()

    def _get_loss_func(
        self,
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Returns the actual function used to compute loss.

        This is overridden by certain classes which compute loss as a side
        effect of training and/or inference.

        Returns:
            Callable[[torch.Tensor, torch.Tensor], torch.Tensor]: configured
                loss function.
        """
        return nn.CrossEntropyLoss(
            ignore_index=special.PAD_IDX,
            label_smoothing=self.label_smoothing,
        )

    def _log_model(self) -> None:
        logging.info("Model: %s", self.name)
        if self.has_features_encoder:
            if self.source_encoder == self.features_encoder:
                logging.info(
                    "Source/features encoder: %s", self.source_encoder.name
                )
            else:
                logging.info("Source encoder: %s", self.source_encoder.name)
                logging.info(
                    "Features encoder: %s", self.features_encoder.name
                )
        else:
            logging.info("Encoder: %s", self.source_encoder.name)
        logging.info("Decoder: %s", self.decoder.name)

    def configure_optimizers(
        self,
    ) -> tuple[list[optim.Optimizer], list[optim.lr_scheduler.LRScheduler]]:
        optimizer = self.optimizer(self.parameters())
        scheduler = self.scheduler(optimizer)
        return [optimizer], [scheduler]

    @abc.abstractmethod
    def get_decoder(self) -> modules.BaseModule: ...

    @staticmethod
    @abc.abstractmethod
    def init_embeddings(
        num_embeddings: int, embedding_size: int
    ) -> nn.Embedding: ...

    # TODO: update with new metrics as they become available.

    @property
    def has_accuracy(self) -> bool:
        return self.accuracy is not None

    @property
    def has_ser(self) -> bool:
        return self.ser is not None

    def start_symbol(self, batch_size: int) -> torch.Tensor:
        """Generates a tensor of start symbols for the batch."""
        return torch.tensor([special.START_IDX], device=self.device).repeat(
            batch_size, 1
        )

    def predict_step(
        self,
        batch: data.Batch,
        batch_idx: int,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Runs one predict step.

        If beam_width > 1 this invokes the forward method directly and returns
        the predictions and scores; otherwise it returns the predictions.

        No metrics are tracked.

        Args:
            batch (data.Batch).
            batch_idx (int).

        Returns:
            Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor].
        """
        if self.beam_width > 1:
            return self(batch)
        else:
            return torch.argmax(self(batch), dim=1)

    def training_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        """Runs one step of training.

        Training loss is tracked.

        Args:
            batch (data.Batch)
            batch_idx (int).

        Returns:
            torch.Tensor: training loss.
        """
        predictions = self(batch)
        loss = self.loss_func(predictions, batch.target.padded)
        self.log(
            "train_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss

    def on_test_epoch_start(self) -> None:
        self._reset_metrics()

    def test_step(self, batch: data.Batch, batch_idx: int) -> None:
        """Runs one test step.

        Evaluation metrics are tracked.

        Args:
            batch (data.Batch).
            batch_idx (int).
        """
        if self.beam_width > 1:
            # TODO: make this workable with `yoyodyne test`.
            predictions, _ = self(batch)
        else:
            predictions = torch.argmax(self(batch), dim=1)
        self._update_metrics(predictions, batch.target.padded)

    def on_test_epoch_end(self) -> None:
        self._log_metrics_on_epoch_end("test")

    def on_validation_epoch_start(self) -> None:
        self._reset_metrics()

    def validation_step(
        self,
        batch: data.Batch,
        batch_idx: int,
    ) -> None:
        """Runs one validation step.

        Validation loss and evaluation metrics are tracked.

        Args:
            batch (data.Batch).
            batch_idx (int).
        """
        predictions = self(batch)
        loss = self.loss_func(predictions, batch.target.padded)
        self.log(
            "val_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            prog_bar=True,
        )
        self._update_metrics(predictions, batch.target.padded)

    def on_validation_epoch_end(self) -> None:
        self._log_metrics_on_epoch_end("val")

    # Helpers to make it run.

    def _reset_metrics(self) -> None:
        # TODO: update with new metrics as they become available.
        if self.has_accuracy:
            self.accuracy.reset()
        if self.has_ser:
            self.ser.reset()

    def _update_metrics(
        self, predictions: torch.Tensor, target: torch.Tensor
    ) -> None:
        # TODO: update with new metrics as they become available.
        if self.has_accuracy:
            self.accuracy.update(predictions, target)
        if self.has_ser:
            self.ser.update(predictions, target)

    def _log_metrics_on_epoch_end(self, subset: str) -> None:
        # TODO: update with new metrics as they become available.
        if self.has_accuracy:
            self.log(
                f"{subset}_accuracy",
                self.accuracy.compute(),
                logger=True,
                on_epoch=True,
                prog_bar=True,
            )
        if self.has_ser:
            self.log(
                f"{subset}_ser",
                self.ser.compute(),
                logger=True,
                on_epoch=True,
                prog_bar=True,
            )



"""Custom callbacks."""

import csv
import itertools
from typing import Optional, Sequence, TextIO, Tuple, Union

import lightning
from lightning.pytorch import callbacks, trainer
import torch

from . import data, defaults, models, util


class PredictionWriter(callbacks.BasePredictionWriter):
    """Writes predictions.

    Args:
        path: Path for the predictions file.
    """

    path: str
    sink: Optional[TextIO]

    # path is given a default argument to silence a warning if no prediction
    # callback is configured.
    # TODO: remove default if this is addressed:
    #
    #   https://github.com/Lightning-AI/pytorch-lightning/issues/20851
    def __init__(
        self,
        path: str = "",
        target_sep: str = defaults.TARGET_SEP,
    ):
        super().__init__("batch")
        self.path = path
        self.sink = None
        self.target_sep = target_sep

    # Required API.

    def on_predict_start(
        self, trainer: trainer.Trainer, pl_module: lightning.LightningModule
    ) -> None:
        # Placing this here prevents the creation of an empty file in the case
        # where a prediction callback was specified but this is not running
        # in predict mode.
        util.mkpath(self.path)
        self.sink = open(self.path, "w")

    def write_on_batch_end(
        self,
        trainer: trainer.Trainer,
        model: models.BaseModel,
        predictions: Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor],
        batch_indices: Sequence[int] | None,
        batch: data.Batch,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        mapper = data.Mapper(trainer.datamodule.index)
        if hasattr(model, "beam_width") and model.beam_width > 1:
            # Beam search.
            tsv_writer = csv.writer(self.sink, delimiter="\t")
            # Even though beam search currently assumes batch size of 1,
            # this assumption is not baked-in here and should generalize
            # if this restriction is lifted.
            for beam, beam_scores in zip(*predictions):
                beam_strings = [
                    self.target_sep.join(mapper.decode_target(prediction))
                    for prediction in beam
                ]
                # Collates target strings and their scores.
                row = itertools.chain.from_iterable(
                    zip(beam_strings, beam_scores.tolist())
                )
                tsv_writer.writerow(row)
        else:
            # Greedy search.
            for prediction in predictions:
                print(
                    self.target_sep.join(mapper.decode_target(prediction)),
                    file=self.sink,
                )
        self.sink.flush()

    def on_predict_end(
        self, trainer: trainer.Trainer, pl_module: lightning.LightningModule
    ) -> None:
        self.sink.close()



"""RNN model classes.

RNNModel is the base class; it has encoder and decoder modules, the classifier,
and the initial decoder hidden state and provides methods for greedy and beam
decoding. Subclassing is used to inject the concrete decoder modules.
"""

import abc
from typing import Optional, Tuple, Union

import torch
from torch import nn

from .. import data, defaults, special
from . import base, beam_search, embeddings, modules


class RNNModel(base.BaseModel):
    """Abstract base class for RNN models.

    The implementation of `get_decoder` in the subclasses determines what kind
    of RNN is used (i.e., GRU or LSTM), and this determines whether the model
    uses a learned attention or not.

    If features are provided, the encodings are fused by concatenation of the
    features encoding with the source encoding on the sequence length
    dimension.

    Args:
        *args: passed to superclass.
        teacher_forcing (bool, optional): should teacher (rather than student)
            forcing be used?
        **kwargs: passed to superclass.
    """

    teacher_forcing: bool
    classifier: nn.Linear

    def __init__(
        self,
        *args,
        teacher_forcing: bool = defaults.TEACHER_FORCING,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.teacher_forcing = teacher_forcing
        self.classifier = nn.Linear(
            self.decoder_hidden_size, self.target_vocab_size
        )
        if (
            self.has_features_encoder
            and self.source_encoder.output_size
            != self.features_encoder.output_size
        ):
            raise base.Error(
                "Cannot concatenate source encoding "
                f"({self.source_encoder.output_size}) and features "
                f"encoding ({self.features_encoder.output_size})"
            )

    def beam_decode(
        self,
        sequence: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Decodes with beam search.

        Decoding halts once all sequences in a batch have reached END. It is
        not currently possible to combine this with loss computation or
        teacher forcing.

        The implementation assumes batch size is 1, but both inputs and outputs
        are still assumed to have a leading dimension representing batch size.

        Args:
            sequence (torch.Tensor).
            encoded (torch.Tensor).
            mask (torch.Tensor).

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: predictions of shape
                B x beam_width x seq_len and log-likelihoods of shape
                B x beam_width.
        """
        # TODO: modify to work with batches larger than 1.
        batch_size = sequence.size(0)
        if batch_size != 1:
            raise NotImplementedError(
                "Beam search is not supported for batch_size > 1"
            )
        state = self.decoder.initial_state(batch_size)
        # The start symbol is not needed here because the beam puts that in
        # automatically.
        beam = beam_search.Beam(self.beam_width, state)
        for _ in range(self.max_target_length):
            for cell in beam.cells:
                if cell.final:
                    beam.push(cell)
                else:
                    symbol = torch.tensor([[cell.symbol]], device=self.device)
                    logits, state = self.decode_step(
                        sequence, encoded, mask, symbol, cell.state
                    )
                    scores = nn.functional.log_softmax(logits.squeeze(), dim=0)
                    for new_cell in cell.extensions(state, scores):
                        beam.push(new_cell)
            beam.update()
            if beam.final:
                break
        return beam.predictions(self.device), beam.scores(self.device)

    def decode_step(
        self,
        sequence: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: modules.RNNState,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Single step of the decoder.

        Args:
            sequence (torch.Tensor).
            encoded (torch.Tensor).
            mask (torch.Tensor).
            symbol (torch.Tensor): next symbol.
            state (modules.RNNState): RNN state.

        Returns:
            Tuple[torch.Tensor, modules.RNNState]: logits and the RNN state.
        """
        decoded, state = self.decoder(
            sequence, encoded, mask, symbol, state, self.embeddings
        )
        logits = self.classifier(decoded)
        return logits, state

    @property
    def decoder_input_size(self) -> int:
        # Features concatenation does not change this.
        return self.source_encoder.output_size

    def forward(
        self,
        batch: data.Batch,
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]: beam search
                returns a tuple with a tensor of predictions of shape
                B x beam_width x seq_len and a tensor of B x shape beam_width
                with the likelihood (the unnormalized sum of sequence
                log-probabilities) for each prediction; greedy search returns
                a tensor of predictions of shape
                B x seq_len x target_vocab_size.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Features column specified but no feature
                encoder specified.
        """
        sequence = batch.source.padded
        encoded = self.source_encoder(batch.source, self.embeddings)
        mask = batch.source.mask
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            sequence = torch.cat((sequence, batch.features.padded), dim=1)
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            encoded = torch.cat((encoded, features_encoded), dim=1)
            mask = torch.cat((mask, batch.features.mask), dim=1)
        elif batch.has_features:
            raise base.ConfigurationError(
                "Features column specified but no feature encoder specified"
            )
        if self.beam_width > 1:
            return self.beam_decode(sequence, encoded, mask)
        else:
            return self.greedy_decode(
                sequence,
                encoded,
                mask,
                self.teacher_forcing if self.training else False,
                batch.target.padded if batch.has_target else None,
            )

    @abc.abstractmethod
    def get_decoder(self) -> modules.RNNDecoder: ...

    def greedy_decode(
        self,
        sequence: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        teacher_forcing: bool = defaults.TEACHER_FORCING,
        target: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Decodes greedily.

        A hard upper bound on the length of the decoded strings is provided by
        the length of the target strings (more specifically, the length of the
        longest target string) if a target is provided, or `max_target_length`
        if not. Decoding will halt earlier if no target is provided and all
        sequences have reached END.

        Args:
            sequence (torch.Tensor).
            encoded (torch.Tensor).
            mask (torch.Tensor).
            teacher_forcing (bool, optional): whether or not to decode with
                teacher forcing.
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues until this length is reached.

        Returns:
            torch.Tensor: predictions of B x target_vocab_size x seq_len.
        """
        batch_size = sequence.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        predictions = []
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        for t in range(max_num_steps):
            logits, state = self.decode_step(
                sequence, encoded, mask, symbol, state
            )
            predictions.append(logits.squeeze(1))
            # With teacher forcing the next input is the gold symbol for this
            # step; with student forcing, it's the top prediction.
            symbol = (
                target[:, t].unsqueeze(1)
                if teacher_forcing
                else torch.argmax(logits, dim=2)
            )
            if target is None:
                # Updates which sequences have decoded an END.
                final = torch.logical_or(final, symbol == special.END_IDX)
                if final.all():
                    break
        predictions = torch.stack(predictions, dim=2)
        return predictions

    def init_embeddings(
        self,
        num_embeddings: int,
        embedding_size: int,
    ) -> nn.Embedding:
        """Initializes the embedding layer.

        Args:
            num_embeddings (int): number of embeddings.
            embedding_size (int): dimension of embeddings.

        Returns:
            nn.Embedding: embedding layer.
        """
        return embeddings.normal_embedding(num_embeddings, embedding_size)

    @property
    @abc.abstractmethod
    def name(self) -> str: ...


class GRUModel(RNNModel):
    """GRU model without attention."""

    def get_decoder(self) -> modules.GRUDecoder:
        return modules.GRUDecoder(
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "GRU"


class LSTMModel(RNNModel):
    """LSTM model without attention.

    Args:
        *args: passed to superclass.
        **kwargs: passed to superclass.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.c0 = nn.Parameter(torch.rand(self.decoder_hidden_size))

    def get_decoder(self) -> modules.LSTMDecoder:
        return modules.LSTMDecoder(
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    def init_state(self, batch_size: int) -> modules.RNNState:
        return modules.RNNState(
            self._init_input(batch_size),
            self._init_hiddens(batch_size),
            self._init_cell(batch_size),
        )

    def _init_cell(self, batch_size: int) -> torch.Tensor:
        return self.c0.repeat(self.decoder_layers, batch_size, 1)

    @property
    def name(self) -> str:
        return "LSTM"


class SoftAttentionGRUModel(GRUModel):
    """GRU model with attention."""

    def get_decoder(self) -> modules.SoftAttentionGRUDecoder:
        return modules.SoftAttentionGRUDecoder(
            attention_input_size=self.decoder_input_size,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "soft attention GRU"


class SoftAttentionLSTMModel(LSTMModel):
    """LSTM model with attention."""

    def get_decoder(self) -> modules.SoftAttentionLSTMDecoder:
        return modules.SoftAttentionLSTMDecoder(
            attention_input_size=self.decoder_input_size,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "soft attention LSTM"



"""Transducer model class."""

import abc
from typing import Callable, Dict, List, Optional, Tuple

from maxwell import actions, sed
import numpy
import torch
from torch import nn

from .. import data, defaults, special, util
from . import base, embeddings, expert, modules


class TransducerRNNModel(base.BaseModel):
    """Abstract base class for transducer models.

    Transducer models are essentially inattentive RNN models which
    predict edits trained using a learned oracle.

    If features are provided, the encodings are fused by concatenation of the
    source encoding with the features encoding, averaged across the length
    dimension and then scattered along the source length dimension, on the
    encoding dimension.

    After:
        Makarov, P., and Clematide, S. 2018. Imitation learning for neural
        morphological string transduction. In Proceedings of the 2018
        Conference on Empirical Methods in Natural Language Processing, pages
        2877–2882.

     Args:
        sed_path (str): path to SED parameters .pkl.
        *args: passed to superclass.
        index (data.Index, optional): index for mapping symbols to indices.
        oracle_factor (int, optional): a scaling factor for scheduling
            predictions during transducer training.
        teacher_forcing (bool, optional): should teacher (rather than student)
            forcing be used?
        **kwargs: passed to superclass.
    """

    expert: expert.Expert
    teacher_forcing: bool
    classifer: nn.Linear

    def __init__(
        self,
        sed_path: str,
        *args,
        index: Optional[data.Index] = None,  # Dummy value filled in via link.
        oracle_factor: int = defaults.ORACLE_FACTOR,
        teacher_forcing: bool = defaults.TEACHER_FORCING,
        **kwargs,
    ):
        self.actions = expert.ActionVocabulary(index)
        self.insertions = self.actions.insertions
        self.substitutions = self.actions.substitutions
        aligner = sed.StochasticEditDistance(
            sed.ParamDict.read_params(sed_path)
        )
        self.expert = expert.Expert(actions, aligner, oracle_factor)
        # The vocabularies are defined in a radically different way here.
        self.vocab_offset = index.vocab_size
        kwargs["target_vocab_size"] = len(self.actions)
        kwargs["vocab_size"] = index.vocab_size + len(self.actions)
        super().__init__(*args, **kwargs)
        self.teacher_forcing = teacher_forcing
        self.classifier = nn.Linear(
            self.decoder_hidden_size, self.target_vocab_size
        )

    def _get_loss_func(
        self,
    ) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        # Prevents base construction of unused loss function.
        return None

    @property
    def decoder_input_size(self) -> int:
        if self.has_features_encoder:
            return (
                self.source_encoder.output_size
                + self.features_encoder.output_size
            )
        else:
            return self.source_encoder.output_size

    def forward(
        self,
        batch: data.Batch,
    ) -> Tuple[List[List[int]], torch.Tensor]:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            Tuple[List[List[int]], torch.Tensor]: encoded prediction values
                and loss tensor; due to transducer setup, prediction is
                performed during training, so these are returned.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Features column specified but no feature
                encoder specified.
        """
        encoded = self.source_encoder(batch.source, self.embeddings)
        # Ignores start symbol.
        encoded = encoded[:, 1:, :]
        source = batch.source.padded[:, 1:]
        source_mask = batch.source.mask[:, 1:]
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            features_encoded = features_encoded.mean(dim=1, keepdim=True)
            features_encoded = features_encoded.expand(-1, encoded.size(1), -1)
            encoded = torch.cat((encoded, features_encoded), dim=2)
        elif batch.has_features:
            raise base.ConfigurationError(
                "Feature column specified but no feature encoder specified"
            )
        return self.greedy_decode(
            source,
            encoded,
            source_mask,
            teacher_forcing=self.teacher_forcing if self.training else False,
            target=batch.target.padded if batch.has_target else None,
            target_mask=batch.target.mask if batch.has_target else None,
        )

    def greedy_decode(
        self,
        source: torch.Tensor,
        encoded: torch.Tensor,
        source_mask: torch.Tensor,
        teacher_forcing: bool,
        target: Optional[torch.Tensor] = None,
        target_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[List[List[int]], torch.Tensor]:
        """Decodes a sequence given the encoded input.

        This essentially serves as a wrapper for looping decode_step.

        Args:
            source (torch.Tensor): source symbols.
            encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for source input.
            teacher_forcing (bool): whether or not to decode
                with teacher forcing; determines whether or not to rollout
                optimal actions.
            target (torch.Tensor, optional): encoded target input.
            target_mask (torch.Tensor, optional): mask for target input.

        Returns:
            Tuple[List[List[int]], torch.Tensor]: encoded prediction values
                and loss tensor; due to transducer setup, prediction is
                performed during training, so these are returned.
        """
        batch_size = source_mask.size(0)
        lengths = (~source_mask).sum(dim=1)
        # Initializing values.
        alignment = torch.zeros(
            batch_size, device=self.device, dtype=torch.int64
        )
        action_count = torch.zeros(
            batch_size, device=self.device, dtype=torch.int64
        )
        last_action = torch.full(
            (batch_size,), self.actions.beg_idx, device=self.device
        )
        loss = torch.zeros(batch_size, device=self.device)
        prediction = [[] for _ in range(batch_size)]
        state = self.decoder.initial_state(batch_size)
        # Converting encodings for prediction.
        if target is not None:
            # Target and source need to be integers for SED values.
            # Clips EOW (idx = -1) for source and target.
            source_list = [
                s[~smask].tolist()[:-1]
                for s, smask in zip(source, source_mask)
            ]
            target_list = [
                t[~tmask].tolist()[:-1]
                for t, tmask in zip(target, target_mask)
            ]
        for _ in range(self.max_target_length):
            # Checks if completed all sequences.
            not_complete = last_action != self.actions.end_idx
            if not any(not_complete):
                break
            # Proceeds to make new edit; new action for all current decoding.
            action_count = torch.where(
                not_complete.to(self.device),
                action_count + 1,
                action_count,
            )
            # Decoding.
            # We offset the action idx by the symbol vocab size so that we
            # can index into the shared embeddings matrix.
            decoded, state = self.decoder(
                source,
                encoded,
                # Accomodates RNNDecoder; see encoder_mask behavior.
                ~(alignment.unsqueeze(1) + 1),
                last_action.unsqueeze(dim=1) + self.vocab_offset,
                state,
                self.embeddings,
            )
            logits = self.classifier(decoded).squeeze(1)
            # If given targets, asks expert for optimal actions.
            optim_actions = (
                self._batch_expert_rollout(
                    source_list,
                    target_list,
                    alignment,
                    prediction,
                    not_complete,
                )
                if target is not None
                else None
            )
            last_action = self._decode_action_step(
                logits,
                alignment,
                lengths,
                not_complete,
                optim_actions=optim_actions if teacher_forcing else None,
            )
            alignment = self._update_prediction(
                last_action, source_list, alignment, prediction
            )
            # If target, validation or training step loss required.
            if target is not None:
                log_sum_loss = self._log_sum_softmax_loss(
                    logits, optim_actions
                )
                loss = torch.where(not_complete, log_sum_loss + loss, loss)
        avg_loss = torch.mean(loss / action_count)
        return prediction, -avg_loss

    def _batch_expert_rollout(
        self,
        source: List[List[int]],
        target: List[List[int]],
        alignment: torch.Tensor,
        prediction: List[List[int]],
        not_complete: torch.Tensor,
    ) -> List[List[int]]:
        """Performs expert rollout over batch."""
        return [
            (
                self._expert_rollout(s, t, align, pred)
                if nc
                else self.actions.end_idx
            )
            for s, t, align, pred, nc in zip(
                source, target, alignment, prediction, not_complete
            )
        ]

    def _expert_rollout(
        self,
        source: List[int],
        target: List[int],
        alignment: int,
        prediction: List[int],
    ) -> List[int]:
        """Rolls out with optimal expert policy.

        Args:
            source (List[int]): input string.
            target (List[int]): target string.
            alignment (int): position in source to edit.
            prediction (List[str]): current prediction.

        Returns:
            List[int]: optimal action encodings.
        """
        raw_action_scores = self.expert.score(
            source,
            target,
            alignment,
            prediction,
            max_action_seq_len=self.max_target_length,
        )
        action_scores = self._remap_actions(raw_action_scores)
        optimal_value = min(action_scores.values())
        optimal_action = sorted(
            [
                self.actions.encode_unseen_action(action)
                for action, value in action_scores.items()
                if value == optimal_value
            ]
        )
        return optimal_action

    def _decode_action_step(
        self,
        logits: torch.Tensor,
        alignment: torch.Tensor,
        lengths: torch.Tensor,
        not_complete: torch.Tensor,
        optim_actions: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Decodes logits to find edit action.

        Finds possible actions given remaining size of source input and masks
        logits for edit action decoding.

        Args:
            logits (torch.Tensor): logit values from decode_step of shape
                B x num_actions.
            alignment (torch.Tensor): index of encoding symbols for decoding,
                per item in batch of shape B x seq_len.
            lengths (torch.Tensor): length of each item in batch.
            not_complete (torch.Tensor): boolean values designating which items
                have not terminated edits.
            optim_actions (List[List[int]], optional): optimal actions
                determined by expert, present when loss is being calculated.

        Returns:
            torch.Tensor: chosen edit action.
        """
        # Finds valid actions given remaining input length.
        end_of_input = (lengths - alignment) <= 1  # 1 -> Last char.
        valid_actions = [
            (
                self._compute_valid_actions(eoi)
                if nc
                else [self.actions.end_idx]
            )
            for eoi, nc in zip(end_of_input, not_complete)
        ]
        # Masks invalid actions.
        logits = self._action_probability_mask(logits, valid_actions)
        return self._choose_action(logits, not_complete, optim_actions)

    def _compute_valid_actions(self, end_of_input: bool) -> List[int]:
        """Gives all possible actions for remaining length of edits.

        Args:
            end_of_input (bool): indicates if this is the last input from
                string; if true, only insertions are available.

        Returns:
            List[actions.Edit]: actions known by transducer.
        """
        valid_actions = [self.actions.end_idx]
        valid_actions.extend(self.insertions)
        if not end_of_input:
            valid_actions.extend([self.actions.copy_idx, self.actions.del_idx])
            valid_actions.extend(self.substitutions)
        return valid_actions

    def _action_probability_mask(
        self, logits: torch.Tensor, valid_actions: List[int]
    ) -> torch.Tensor:
        """Masks non-valid actions in logits."""
        with torch.no_grad():
            mask = torch.full(
                logits.shape, defaults.NEG_INF, device=self.device
            )
            for row, action in zip(mask, valid_actions):
                row[action] = 0.0
            logits = mask + logits
        return logits

    def _choose_action(
        self,
        logits: torch.Tensor,
        not_complete: torch.Tensor,
        optim_actions: Optional[List[List[int]]] = None,
    ) -> torch.Tensor:
        """Chooses transducer action from log_prob distribution.

        If training, uses dynamic oracle for selection.

        Args:
            log_probs (torch.Tensor): probability distribution of actions.
            not_complete (torch.Tensor): boolean tensor of batch length to
                indicate if each item in batch is complete.
            optim_actions (Optional[List[List[int]]]): optional encoded actions
                to use for action selection.

        Returns:
            torch.Tensor: action encodings.
        """
        # TODO: Merge logic into PyTorch methods.
        log_probs = nn.functional.log_softmax(logits, dim=1)
        if optim_actions is None:
            # Argmax decoding.
            next_action = [
                (torch.argmax(probs, dim=0) if nc else self.actions.end_idx)
                for probs, nc in zip(log_probs, not_complete)
            ]
        else:
            # Training with dynamic oracle; chooses from optimal actions.
            with torch.no_grad():
                if self.expert.explore():
                    # Action is picked by random exploration.
                    next_action = [
                        (self._sample(probs) if nc else self.actions.end_idx)
                        for probs, nc in zip(log_probs, not_complete)
                    ]
                else:
                    # Action is picked from optim_actions.
                    next_action = []
                    for action, probs, nc in zip(
                        optim_actions, log_probs, not_complete
                    ):
                        if nc:
                            optim_logs = probs[action]
                            idx = int(torch.argmax(optim_logs, 0))
                            next_action.append(action[idx])
                        else:  # Already complete, so skip.
                            next_action.append(self.actions.end_idx)
        return torch.tensor(next_action, device=self.device, dtype=torch.int)

    @staticmethod
    def _sample(log_probs: torch.Tensor) -> torch.Tensor:
        """Samples an action from a log-probability distribution."""
        dist = torch.exp(log_probs)
        rand = numpy.random.rand()
        for action, p in enumerate(dist):
            rand -= p
            if rand <= 0:
                break
        return action

    # TODO: Merge action classes to remove need for this method.
    @staticmethod
    def _remap_actions(
        action_scores: Dict[actions.Edit, float],
    ) -> Dict[actions.Edit, float]:
        """Maps generative oracle's edit to conditional counterpart.

        Oracle edits are a distinct subclass from edits learned from samples.

        This will eventually be removed.

        Args:
            action_scores (Dict[actions.Edit, float]): weights for each action.

        Returns:
            Dict[actions.Edit, float]: edit action-weight pairs.
        """
        remapped_action_scores = {}
        for action, score in action_scores.items():
            if isinstance(action, actions.GenerativeEdit):
                remapped_action = action.conditional_counterpart()
            elif isinstance(action, actions.Edit):
                remapped_action = action
            else:
                raise expert.ActionError(
                    f"Unknown action: {action}, {score}, "
                    f"action_scores: {action_scores}"
                )
            remapped_action_scores[remapped_action] = score
        return remapped_action_scores

    def _update_prediction(
        self,
        action: List[actions.Edit],
        source: List[int],
        alignment: torch.Tensor,
        prediction: List[List[str]],
    ) -> torch.Tensor:
        """Batch updates prediction and alignment information from actions.

        Args:
           action (List[actions.Edit]): valid actions, one per item in
                batch.
           source (List[int]): source strings, one per item in batch.
           alignment (torch.Tensor): index of current symbol for each item
                in batch.
           prediction (List[List[str]]): current predictions for each item
                in batch, one list of symbols per item.

        Return:
            torch.Tensor: new alignments for transduction.
        """
        alignment_update = torch.zeros(
            alignment.shape, device=self.device, dtype=torch.int64
        )
        for i in range(len(source)):
            a = self.actions.decode(action[i])
            if isinstance(a, actions.ConditionalCopy):
                symb = source[i][alignment[i]]
                prediction[i].append(symb)
                alignment_update[i] += 1
            elif isinstance(a, actions.ConditionalDel):
                alignment_update[i] += 1
            elif isinstance(a, actions.ConditionalIns):
                prediction[i].append(a.new)
            elif isinstance(a, actions.ConditionalSub):
                alignment_update[i] += 1
                prediction[i].append(a.new)
            elif isinstance(a, actions.End):
                prediction[i].append(special.END_IDX)
            else:
                raise expert.ActionError(f"Unknown action: {action[i]}")
        return alignment + alignment_update

    @staticmethod
    def _log_sum_softmax_loss(
        logits: torch.Tensor, optimal_actions: List[int]
    ) -> torch.Tensor:
        """Computes log loss.

        After:
            Riezler, S., Prescher, D., Kuhn, J., and Johnson, M. 2000.
            Lexicalized stochastic modeling of constraint-based grammars using
            log-linear measures and EM training. In Proceedings of the 38th
            Annual Meeting of the Association for Computational
            Linguistics, pages 480–487.
        """
        opt_act = [
            log[actions] for log, actions in zip(logits, optimal_actions)
        ]
        log_sum_exp_terms = torch.stack(
            [torch.logsumexp(act, dim=-1) for act in opt_act]
        )
        normalization_term = torch.logsumexp(logits, -1)
        return log_sum_exp_terms - normalization_term

    def predict_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        predictions, _ = self(batch)
        length = max(len(prediction) for prediction in predictions)
        # Pads; truncation cannot occur by construction.
        return self._convert_predictions(predictions, length)

    def test_step(self, batch: data.Batch, batch_idx: int) -> None:
        predictions, _ = self(batch)
        self._update_metrics(
            self._convert_predictions(predictions), batch.target.padded
        )

    def on_train_epoch_start(self) -> None:
        self.expert.roll_in_schedule(self.current_epoch)

    def training_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        """Runs one step of training.

        Training loss is tracked.

        Args:
            batch (data.Batch)
            batch_idx (int).

        Returns:
            torch.Tensor: training loss.
        """
        # Forward pass produces loss.
        _, loss = self(batch)
        self.log(
            "train_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: data.Batch, batch_idx: int) -> None:
        predictions, loss = self(batch)
        self.log(
            "val_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            prog_bar=True,
        )
        # This needs to conform to target size for evaluation.
        length = batch.target.padded.size(1)
        self._update_metrics(
            self._convert_predictions(predictions, length),
            batch.target.padded,
        )

    def _convert_predictions(
        self, predictions: List[List[int]], length: int
    ) -> torch.Tensor:
        """Converts a batch of predictions to the proper form.

        This repeatedly calls `_resize_prediction`, stacks, and then converts
        redundant END to PAD.

        Args:
            predictions (list[list[int][): lists of prediction indices.
            length (int): desired length.

        Returns:
            torch.Tensor.
        """
        return util.pad_tensor_after_end(
            torch.stack(
                [
                    self._resize_prediction(prediction, length)
                    for prediction in predictions
                ]
            )
        )

    def _resize_prediction(
        self, prediction: List[int], length: int
    ) -> torch.Tensor:
        """Resizes the prediction and converts to tensor.

        If the prediction matches the desired length it is just converted to
        tensor. If the prediction is longer than the desired length, it is
        first truncated. If the prediction is shorter than the desired length,
        it is padded using END.

        Args:
            predictions (list[int]): prediction indices.
            length (int): desired length.

        Returns:
            torch.Tensor.
        """
        if len(prediction) == length:
            # Just converts to tensor.
            return torch.tensor(prediction, device=self.device)
        elif len(prediction) < length:
            # Pads.
            padding = length - len(prediction)
            return nn.functional.pad(
                torch.tensor(prediction, device=self.device),
                (0, padding),
                "constant",
                self.actions.end_idx,
            )
        else:
            # Truncates; this is never used during the prediction step hence
            # its late ordering.
            return torch.tensor(prediction[:length], device=self.device)

    def init_embeddings(
        self,
        num_embeddings: int,
        embedding_size: int,
    ) -> nn.Embedding:
        """Initializes the embedding layer.

        Args:
            num_embeddings (int): number of embeddings.
            embedding_size (int): dimension of embeddings.

        Returns:
            nn.Embedding: embedding layer.
        """
        return embeddings.normal_embedding(num_embeddings, embedding_size)

    @property
    @abc.abstractmethod
    def name(self) -> str: ...


class TransducerGRUModel(TransducerRNNModel):
    """Transducer with GRU backend."""

    def get_decoder(self) -> modules.GRUDecoder:
        return modules.GRUDecoder(
            bidirectional=False,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embeddings=self.embeddings,
            embedding_size=self.embedding_size,
            layers=self.decoder_layers,
            hidden_size=self.decoder_hidden_size,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "transducer GRU"


class TransducerLSTMModel(TransducerRNNModel):
    """Transducer with LSTM backend."""

    def get_decoder(self) -> modules.LSTMDecoder:
        return modules.LSTMDecoder(
            bidirectional=False,
            decoder_input_size=self.decoder_input_size,
            dropout=self.decoder_dropout,
            embeddings=self.embeddings,
            embedding_size=self.embedding_size,
            layers=self.decoder_layers,
            hidden_size=self.decoder_hidden_size,
            num_embeddings=self.num_embeddings,
        )

    @property
    def name(self) -> str:
        return "transducer LSTM"



"""Data modules."""

import logging
from typing import Iterable, Optional, Set

import lightning
from torch.utils import data

from .. import defaults
from . import collators, datasets, indexes, mappers, tsv


class DataModule(lightning.LightningDataModule):
    """Data module.

    This is responsible for indexing the data, collating/padding, and
    generating datasets.

    Args:
        model_dir: Path for checkpoints, indexes, and logs.
        train: Path for training data TSV.
        val: Path for validation data TSV.
        predict: Path for prediction data TSV.
        test: Path for test data TSV.
        source_col: 1-indexed column in TSV containing source strings.
        features_col: 1-indexed column in TSV containing features strings.
        target_col: 1-indexed column in TSV containing target strings.
        source_sep: String used to split source string into symbols; an empty
            string indicates that each Unicode codepoint is its own symbol.
        features_sep: String used to split features string into symbols; an
            empty string indicates that each Unicode codepoint is its own
            symbol.
        target_sep: String used to split target string into symbols; an empty
            string indicates that each Unicode codepoint is its own symbol.
        tie_embeddings: Whether or not source and target embeddings are tied.
            If not, then source symbols are wrapped in {...}.
        batch_size: Desired batch size.
        max_source_length: The maximum length of a source string. An error will
            be raised if any source string exceeds this limit.
        max_features_length: The maximum length of a features string. An error
            will be raised if any features string exceeds this limit.
        max_target_length: The maximum length of a target string. A warning
            will be raised and the strings will be truncated if any target
            exceeds this limit.
    """

    train: Optional[str]
    val: Optional[str]
    predict: Optional[str]
    test: Optional[str]
    parser: tsv.TsvParser
    batch_size: int
    index: indexes.Index
    collator: collators.Collator

    def __init__(
        self,
        # Paths.
        *,
        model_dir: str,
        train: Optional[str] = None,
        val: Optional[str] = None,
        predict: Optional[str] = None,
        test: Optional[str] = None,
        # TSV parsing arguments.
        source_col: int = defaults.SOURCE_COL,
        features_col: int = defaults.FEATURES_COL,
        target_col: int = defaults.TARGET_COL,
        source_sep: str = defaults.SOURCE_SEP,
        features_sep: str = defaults.FEATURES_SEP,
        target_sep: str = defaults.TARGET_SEP,
        tie_embeddings: bool = defaults.TIE_EMBEDDINGS,
        # Other.
        batch_size: int = defaults.BATCH_SIZE,
        max_source_length: int = defaults.MAX_LENGTH,
        max_features_length: int = defaults.MAX_LENGTH,
        max_target_length: int = defaults.MAX_LENGTH,
    ):
        super().__init__()
        self.model_dir = model_dir
        self.train = train
        self.val = val
        self.predict = predict
        self.test = test
        self.parser = tsv.TsvParser(
            source_col=source_col,
            features_col=features_col,
            target_col=target_col,
            source_sep=source_sep,
            features_sep=features_sep,
            target_sep=target_sep,
            tie_embeddings=tie_embeddings,
        )
        self.batch_size = batch_size
        # If the training data is specified, it is used to create (or recreate)
        # the index; if not specified it is read from the model directory.
        self.index = (
            self._make_index(tie_embeddings)
            if self.train
            else indexes.Index.read(self.model_dir)
        )
        self.log_vocabularies()
        self.collator = collators.Collator(
            has_features=self.has_features,
            has_target=self.has_target,
            max_source_length=max_source_length,
            max_features_length=max_features_length,
            max_target_length=max_target_length,
        )

    def _make_index(self, tie_embeddings: bool) -> indexes.Index:
        """Creates the index from a training set."""
        source_vocabulary: Set[str] = set()
        features_vocabulary: Set[str] = set()
        target_vocabulary: Set[str] = set()
        if self.has_features:
            if self.has_target:
                for source, features, target in self.parser.samples(
                    self.train
                ):
                    source_vocabulary.update(source)
                    features_vocabulary.update(features)
                    target_vocabulary.update(target)
            else:
                for source, features in self.parser.samples(self.train):
                    source_vocabulary.update(source)
                    features_vocabulary.update(features)
        elif self.has_target:
            for source, target in self.parser.samples(self.train):
                source_vocabulary.update(source)
                target_vocabulary.update(target)
        else:
            for source in self.parser.samples(self.train):
                source_vocabulary.update(source)
        index = indexes.Index(
            source_vocabulary=source_vocabulary,
            features_vocabulary=(
                features_vocabulary if features_vocabulary else None
            ),
            target_vocabulary=(
                target_vocabulary if target_vocabulary else None
            ),
            tie_embeddings=tie_embeddings,
        )
        # Writes it to the model directory.
        index.write(self.model_dir)
        return index

    # Logging.

    @staticmethod
    def pprint(vocabulary: Iterable) -> str:
        """Prints the vocabulary for debugging dnd logging purposes."""
        return ", ".join(f"{symbol!r}" for symbol in vocabulary)

    def log_vocabularies(self) -> None:
        """Logs this module's vocabularies."""
        logging.info(
            "Source vocabulary: %s", self.pprint(self.index.source_vocabulary)
        )
        if self.has_features:
            logging.info(
                "Features vocabulary: %s",
                self.pprint(self.index.features_vocabulary),
            )
        if self.has_target:
            logging.info(
                "Target vocabulary: %s",
                self.pprint(self.index.target_vocabulary),
            )

    # Properties.

    # has_source is always true.

    @property
    def has_features(self) -> bool:
        return self.parser.has_features

    @property
    def has_target(self) -> bool:
        return self.parser.has_target

    @property
    def target_sep(self) -> str:
        return self.parser.target_sep

    @property
    def target_vocab_size(self) -> int:
        return self.index.target_vocab_size

    @property
    def vocab_size(self) -> int:
        return self.index.vocab_size

    # Required API.

    def train_dataloader(self) -> data.DataLoader:
        assert self.train is not None, "no train path"
        return data.DataLoader(
            self._dataset(self.train),
            collate_fn=self.collator,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=1,
            persistent_workers=True,
        )

    def val_dataloader(self) -> data.DataLoader:
        assert self.val is not None, "no val path"
        return data.DataLoader(
            self._dataset(self.val),
            collate_fn=self.collator,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=1,
            persistent_workers=True,
        )

    def predict_dataloader(self) -> data.DataLoader:
        assert self.predict is not None, "no predict path"
        return data.DataLoader(
            self._dataset(self.predict),
            collate_fn=self.collator,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=1,
            persistent_workers=True,
        )

    def test_dataloader(self) -> data.DataLoader:
        assert self.test is not None, "no test path"
        return data.DataLoader(
            self._dataset(self.test),
            collate_fn=self.collator,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=1,
            persistent_workers=True,
        )

    def _dataset(self, path: str) -> datasets.Dataset:
        return datasets.Dataset(
            list(self.parser.samples(path)),
            mappers.Mapper(self.index),
            self.parser,
        )



"""RNN module classes.

RNNModule is the base class.

We abstract away from the different formats used by GRUs and LSTMs; the latter
also tracks "cell state" and stores it as part of a tuple of tensors along
with the hidden state. RNNState hides this detail. For encoder modules,
WrappedGRUEncoderWrapped and WrappedLSTMEncoder wrap nn.GRU and nn.LSTM,
respectively, taking responsibility for packing and padding; WrappedGRUDecoder
and WrappedLSTMDecoder are similar wrappers for decoder modules.
"""

import abc
from typing import Optional, Tuple

import torch
from torch import nn

from ... import data, defaults, special
from . import attention, base


class RNNModule(base.BaseModule):
    """Abstract base class for RNN modules.

    Args:
        *args: passed to superclass.
        hidden_size (int, optional): size of the hidden layer.
        layers (int, optional): number of layers.
        **kwargs: passed to superclass.
    """

    hidden_size: int
    layers: int

    def __init__(
        self,
        *args,
        hidden_size: int = defaults.HIDDEN_SIZE,
        layers: int = defaults.LAYERS,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.hidden_size = hidden_size
        self.layers = layers
        self.module = self.get_module()

    @abc.abstractmethod
    def get_module(self) -> nn.RNNBase: ...

    def embed(
        self, symbols: torch.Tensor, embeddings: nn.Embedding
    ) -> torch.Tensor:
        return self.dropout_layer(embeddings(symbols))


class RNNState(nn.Module):
    """Represents the state of an RNN."""

    hidden: torch.Tensor
    cell: Optional[torch.Tensor]  # LSTMs only.

    def __init__(self, hidden, cell=None):
        super().__init__()
        self.register_buffer("hidden", hidden)
        self.register_buffer("cell", cell)


class WrappedRNNEncoder:
    """Wraps RNN encoder modules to work with packing.

    The derived modules do not pass an initial hidden state (or cell state, in
    the case of LSTMs), so it is effectively zero.
    """

    @staticmethod
    def _pad(sequence: nn.utils.rnn.PackedSequence) -> torch.Tensor:
        padded, _ = nn.utils.rnn.pad_packed_sequence(
            sequence,
            batch_first=True,
            padding_value=special.PAD_IDX,
        )
        return padded

    @staticmethod
    def _pack(
        sequence: torch.Tensor, lengths: torch.Tensor
    ) -> nn.utils.rnn.PackedSequence:
        return nn.utils.rnn.pack_padded_sequence(
            sequence,
            lengths,
            batch_first=True,
            enforce_sorted=False,
        )

    @abc.abstractmethod
    def forward(
        self, sequence: torch.Tensor, lengths: torch.Tensor
    ) -> torch.Tensor: ...


class WrappedGRUEncoder(nn.GRU, WrappedRNNEncoder):
    """Wraps GRU API to work with packing."""

    def forward(
        self,
        sequence: torch.Tensor,
        lengths: torch.Tensor,
    ) -> torch.Tensor:
        packed, _ = super().forward(self._pack(sequence, lengths))
        return self._pad(packed)


class WrappedLSTMEncoder(nn.LSTM, WrappedRNNEncoder):
    """Wraps LSTM API to work with packing."""

    def forward(
        self, sequence: torch.Tensor, lengths: torch.Tensor
    ) -> torch.Tensor:
        packed, _ = super().forward(self._pack(sequence, lengths))
        return self._pad(packed)


class RNNEncoder(RNNModule):
    """Abstract base class for RNN encoders.

    Args:
        *args: passed to superclass.
        bidirectional (bool, optional): should the RNN be bidirectional?
        **kwargs: passed to superclass.
    """

    def __init__(self, *args, bidirectional=defaults.BIDIRECTIONAL, **kwargs):
        self.bidirectional = bidirectional
        super().__init__(*args, **kwargs)

    def forward(
        self, source: data.PaddedTensor, embeddings: nn.Embedding
    ) -> torch.Tensor:
        """Encodes the source.

        Subsequent operations don't use the encoder's final hidden state for
        anything so it's discarded here.

        Args:
            source (data.PaddedTensor): source padded tensors of shape
                B x seq_len x 1.
            embeddings (nn.Embedding): embeddings.

        Returns:
            torch.Tensor.
        """
        return self.module(
            self.embed(source.padded, embeddings), source.lengths()
        )

    @property
    def num_directions(self) -> int:
        return 2 if self.bidirectional else 1

    @property
    def output_size(self) -> int:
        return self.hidden_size * self.num_directions


class GRUEncoder(RNNEncoder):
    """GRU encoder."""

    def get_module(self) -> WrappedGRUEncoder:
        return WrappedGRUEncoder(
            self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=self.bidirectional,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    @property
    def name(self) -> str:
        return "GRU"


class LSTMEncoder(RNNEncoder):
    """LSTM encoder."""

    def get_module(self) -> WrappedLSTMEncoder:
        return WrappedLSTMEncoder(
            self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=self.bidirectional,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    @property
    def name(self) -> str:
        return "LSTM"


class WrappedGRUDecoder(nn.GRU):
    """Wraps GRU API to work with RNNState."""

    def forward(
        self, symbol: torch.Tensor, state: RNNState
    ) -> Tuple[torch.Tensor, RNNState]:
        decoded, hidden = super().forward(symbol, state.hidden)
        return decoded, RNNState(hidden)


class WrappedLSTMDecoder(nn.LSTM):
    """Wraps LSTM API to work with RNNState."""

    def forward(
        self, symbol: torch.Tensor, state: RNNState
    ) -> Tuple[torch.Tensor, RNNState]:
        assert state.cell is not None, "expected cell state"
        decoded, (hidden, cell) = super().forward(
            symbol, (state.hidden, state.cell)
        )
        return decoded, RNNState(hidden, cell)


class RNNDecoder(RNNModule):
    """Abstract base class for RNN decoders.

    This implementation lacks a learned attention mechanism; rather, it uses
    the encodings of the sequence-final hidden states as the input to the
    decoder.

    The initial decoder hidden state is a learned parameter.

    Args:
        *args: passed to superclass.
        decoder_input_size (int): decoder input size.
        **kwargs: passed to superclass.
    """

    decoder_input_size: int
    h0: nn.Parameter

    def __init__(self, decoder_input_size, *args, **kwargs):
        self.decoder_input_size = decoder_input_size
        super().__init__(*args, **kwargs)
        self.h0 = nn.Parameter(torch.rand(self.hidden_size))

    def forward(
        self,
        sequence: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: RNNState,
        embeddings: nn.Embedding,
    ) -> Tuple[torch.Tensor, RNNState]:
        """Single decode pass.

        Args:
            sequence (torch.Tensor): sequence of shape B x seq_len.
            encoded (torch.Tensor): encoded source sequence of shape
                B x seq_len x encoder_dim.
            mask (torch.Tensor): ignored.
            symbol (torch.Tensor): previously decoded symbol(s) of shape B x 1.
            state (RNNState): RNN state.
            embeddings (nn.Embedding): embeddings.

        Returns:
            Tuple[torch.Tensor, RNNState].
        """
        embedded = self.embed(symbol, embeddings)
        end_mask = (
            (sequence == special.END_IDX)
            .unsqueeze(2)
            .expand(-1, -1, encoded.size(2))
        )
        ends = torch.sum(encoded * end_mask, dim=1, keepdim=True)
        decoded, state = self.module(torch.cat((embedded, ends), dim=2), state)
        return self.dropout_layer(decoded), state

    @abc.abstractmethod
    def initial_state(self, batch_size: int) -> RNNState: ...

    @property
    def output_size(self) -> int:
        return self.hidden_size


class GRUDecoder(RNNDecoder):
    """GRU decoder."""

    def get_module(self) -> WrappedGRUDecoder:
        return WrappedGRUDecoder(
            self.decoder_input_size + self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=False,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    def initial_state(self, batch_size: int) -> RNNState:
        return RNNState(self.h0.repeat(self.layers, batch_size, 1))

    @property
    def name(self) -> str:
        return "GRU"


class LSTMDecoder(RNNDecoder):
    """LSTM decoder.

    This also has an initial cell state parameter.
    """

    c0: nn.Parameter

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.c0 = nn.Parameter(torch.rand(self.hidden_size))

    def get_module(self) -> WrappedLSTMDecoder:
        return WrappedLSTMDecoder(
            self.decoder_input_size + self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=False,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    def initial_state(self, batch_size: int) -> RNNState:
        return RNNState(
            self.h0.repeat(self.layers, batch_size, 1),
            self.c0.repeat(self.layers, batch_size, 1),
        )

    @property
    def name(self) -> str:
        return "LSTM"


class SoftAttentionRNNDecoder(RNNDecoder):
    """Abstract base class for soft attention RNN decoders.

    The attention module learns to differentially attend to different symbols
    in the encoder output.
    """

    def __init__(self, attention_input_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attention = attention.Attention(
            attention_input_size, self.hidden_size
        )

    def forward(
        self,
        sequence: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: RNNState,
        embeddings: nn.Embedding,
    ) -> Tuple[torch.Tensor, RNNState]:
        """Single decode pass.

        Args:
            sequence (torch.Tensor): ignored.
            encoded (torch.Tensor): encoded source sequence of shape
                B x seq_len x encoder_dim.
            mask (torch.Tensor): mask of shape B x seq_len.
            symbol (torch.Tensor): previously decoded symbol(s) of shape
                B x 1.
            state (RNNState): RNN state.
            embeddings (nn.Embedding): embeddings.

        Returns:
            Tuple[torch.Tensor, RNNState].
        """
        embedded = self.embed(symbol, embeddings)
        context, _ = self.attention(
            encoded, state.hidden.transpose(0, 1), mask
        )
        decoded, state = self.module(
            torch.cat((embedded, context), dim=2), state
        )
        return self.dropout_layer(decoded), state


class SoftAttentionGRUDecoder(SoftAttentionRNNDecoder, GRUDecoder):
    """Soft attention GRU decoder."""

    @property
    def name(self) -> str:
        return "soft attention GRU"


class SoftAttentionLSTMDecoder(SoftAttentionRNNDecoder, LSTMDecoder):
    """Soft attention LSTM decoder."""

    @property
    def name(self) -> str:
        return "soft attention LSTM"



"""Hard monotonic neural HMM classes."""

import abc
from typing import Callable, Optional, Tuple

import torch
from torch import nn

from .. import data, defaults, special
from . import base, embeddings, modules


class HardAttentionRNNModel(base.BaseModel):
    """Abstract base class for hard attention models.

    Learns probability distribution of target string by modeling transduction
    of source string to target string as Markov process. Assumes each symbol
    produced is conditioned by state transitions over each source symbol.

    The default model assumes independence between state and non-monotonic
    progression over source string. `enforce_monotonic` enforces monotonic
    state transition (model progresses over each source symbol), and
    attention context allows conditioning of state transition over the
    previous n states.

    If features are provided, the encodings are fused by concatenation of the
    source encoding with the features encoding, averaged across the length
    dimension and then scattered along the source length dimension, on the
    encoding dimension.

    After:
        Wu, S. and Cotterell, R. 2019. Exact hard monotonic attention for
        symbol-level transduction. In _Proceedings of the 57th Annual
        Meeting of the Association for Computational Linguistics_, pages
        1530-1537.

    Original implementation:
        https://github.com/shijie-wu/neural-transducer

     Args:
        *args: passed to superclass.
        attention_context (int, optional): size of context window for
            conditioning state transition; if 0, state transitions are
            independent.
        enforce_monotonic (bool, optional): enforces monotonic state
            transition in decoding.
        **kwargs: passed to superclass.
    """

    enforce_monotonic: bool
    attention_context: int

    def __init__(
        self,
        *args,
        attention_context: int = defaults.ATTENTION_CONTEXT,
        enforce_monotonic: bool = defaults.ENFORCE_MONOTONIC,
        **kwargs,
    ):
        self.attention_context = attention_context
        self.enforce_monotonic = enforce_monotonic
        super().__init__(*args, **kwargs)
        self.classifier = nn.Linear(
            self.decoder.output_size, self.target_vocab_size
        )

    def _get_loss_func(
        self,
    ) -> Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]:
        """Returns the actual function used to compute loss.

        This overrides the inherited loss function.

        Returns:
            Callable[[torch.Tensor, torch.Tensor], torch.Tensor]: configured
                loss function.
        """
        return self._loss

    def _loss(
        self,
        target: torch.Tensor,
        emissions: torch.Tensor,
        transitions: torch.Tensor,
    ) -> torch.Tensor:
        # TODO: Currently we're storing a concatenation of loss tensors for
        # each time step. This is costly. Revisit this calculation and see if
        # we can use DP to simplify.
        fwd = transitions[0, :, 0].unsqueeze(1) + self._gather_at_idx(
            emissions[0], target[:, 0]
        )
        for idx in range(1, target.size(1)):
            fwd = fwd + transitions[idx].transpose(1, 2)
            fwd = fwd.logsumexp(dim=2, keepdim=True).transpose(1, 2)
            fwd = fwd + self._gather_at_idx(emissions[idx], target[:, idx])
        loss = -torch.logsumexp(fwd, dim=2).mean() / target.size(1)
        return loss

    def decode(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        target: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Decodes a sequence given the encoded input.

        Decodes until all sequences in a batch have reached END or the
        maximum length of the target, whichever comes first.

        Args:
            encoded (torch.Tensor): encoded source symbols
                of shape B x src_len x (encoder_hidden * num_directions).
            mask (torch.Tensor): mask.
            target (torch.Tensor): target symbols.

        Returns:
            Tuple[torch.Tensor,torch.Tensor]: emission probabilities
                of shape target_len x B x src_len x vocab_size, and
                transition probabilities of shape
                target_len x B x src_len x src_len.
        """
        batch_size = mask.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        emissions, transitions, state = self.decode_step(
            encoded, mask, symbol, state
        )
        all_emissions = [emissions]
        all_transitions = [transitions]
        for idx in range(target.size(1)):
            symbol = target[:, idx].unsqueeze(1)
            emissions, transitions, state = self.decode_step(
                encoded,
                mask,
                symbol,
                state,
            )
            all_emissions.append(emissions)
            all_transitions.append(transitions)
        return torch.stack(all_emissions), torch.stack(all_transitions)

    def decode_step(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: modules.RNNState,
    ) -> Tuple[torch.Tensor, torch.Tensor, modules.RNNState]:
        """Single decoder step.

        Args:
            encoded (torch.Tensor): encoded source symbols of shape
                B x src_len x (encoder_hidden * num_directions).
            mask (torch.Tensor): mask.
            symbol (torch.Tensor): target symbol for current state.
            state (modules.RNNState): RNN state.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: emission
                probabilities, transition probabilities, and the RNN state.
        """
        emissions, transitions, state = self.decoder(
            encoded,
            mask,
            symbol,
            state,
            self.embeddings,
        )
        logits = self.classifier(emissions)
        scores = nn.functional.log_softmax(logits, dim=2)
        # Expands matrix for all time steps.
        if self.enforce_monotonic:
            transitions = self._apply_mono_mask(transitions)
        return scores, transitions, state

    def greedy_decode(
        self,
        source_encoded: torch.Tensor,
        source_mask: torch.Tensor,
        target: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Decodes a sequence given the encoded input.

        Decodes until all sequences in a batch have reached END up to a
        specified length depending on the `target` args.

        Args:
            source_encoded (torch.Tensor): encoded source symbols of shape
                B x src_len x (encoder_hidden * num_directions).
            source_mask (torch.Tensor): mask.
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues until this length is reached.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: predictions of shape
                B x pred_seq_len and per-step likelihoods of shape
                B x 1 x src_len.
        """
        batch_size = source_mask.size(0)
        symbol = self.start_symbol(batch_size)
        state = self.decoder.initial_state(batch_size)
        emissions, transitions, state = self.decode_step(
            source_encoded, source_mask, symbol, state
        )
        likelihood = transitions[:, 0].unsqueeze(1)
        symbol = self._greedy_step(emissions, likelihood)
        predictions = [symbol]
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        # We already did one step.
        for _ in range(max_num_steps - 1):
            emissions, transitions, state = self.decode_step(
                source_encoded,
                source_mask,
                symbol.unsqueeze(1),
                state,
            )
            likelihood = likelihood + transitions.transpose(1, 2)
            likelihood = likelihood.logsumexp(dim=2, keepdim=True).transpose(
                1, 2
            )
            symbol = self._greedy_step(emissions, likelihood)
            predictions.append(symbol)
            if target is None:
                final = torch.logical_or(final, symbol == special.END_IDX)
                if final.all():
                    break
            likelihood = likelihood + self._gather_at_idx(emissions, symbol)
        # -> B x seq_len.
        predictions = torch.stack(predictions, dim=1)
        return predictions, likelihood

    @staticmethod
    def _greedy_step(
        emissions: torch.Tensor, likelihood: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Greedy decoding of current timestep.

        Args:
            emissions (torch.Tensor): emission probabilities at current
                time step.
            likelihood (torch.Tensor): accumulative likelihood of decoded
                symbol sequence.

        Returns:
            torch.Tensor: greedily decoded symbol for the current timestep.
        """
        probabilities = likelihood + emissions.transpose(1, 2)
        probabilities = probabilities.logsumexp(dim=2)
        # -> B.
        return torch.argmax(probabilities, dim=1)

    @staticmethod
    def _gather_at_idx(
        emissions: torch.Tensor, symbol: torch.Tensor
    ) -> torch.Tensor:
        """Collects probability of the symbol across all states.

        To calculate the final emission probability, the pseudo-HMM
        graph needs to aggregate the final emission probabilities of
        target symbols across all potential hidden states in the emissions.

        Args:
            emissions (torch.Tensor): log probabilities of emission states of
                shape B x src_len.
            symbol (torch.Tensor): target symbol to poll probabilities for
                shape B.

        Returns:
            torch.Tensor: emission probabilities of the symbol for each hidden
                state, of size B 1 x src_len.
        """
        batch_size = emissions.size(0)
        src_seq_len = emissions.size(1)
        idx = symbol.view(-1, 1).expand(batch_size, src_seq_len).unsqueeze(-1)
        output = torch.gather(emissions, -1, idx).view(
            batch_size, 1, src_seq_len
        )
        idx = idx.view(batch_size, 1, src_seq_len)
        pad_mask = (idx != special.PAD_IDX).float()
        return output * pad_mask

    def forward(
        self,
        batch: data.Batch,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            Tuple[torch.Tensor,torch.Tensor]: emission probabilities for
                each transition state of shape target_len x B x src_len
                x vocab_size, and transition probabilities for each transition

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Feature column specified but no feature
                encoder specified.
            NotImplementedError: Beam search not implemented.
        """
        encoded = self.source_encoder(batch.source, self.embeddings)
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            features_encoded = features_encoded.mean(dim=1, keepdim=True)
            features_encoded = features_encoded.expand(-1, encoded.size(1), -1)
            encoded = torch.cat((encoded, features_encoded), dim=2)
        elif batch.has_features:
            raise base.ConfigurationError(
                "Feature column specified but no feature encoder specified"
            )
        if self.training:
            return self.decode(
                encoded,
                batch.source.mask,
                batch.target.padded,
            )
        else:
            return self.greedy_decode(
                encoded,
                batch.source.mask,
                batch.target.padded if batch.has_target else None,
            )

    def init_embeddings(
        self,
        num_embeddings: int,
        embedding_size: int,
    ) -> nn.Embedding:
        """Initializes the embedding layer.

        Args:
            num_embeddings (int): number of embeddings.
            embedding_size (int): dimension of embeddings.

        Returns:
            nn.Embedding: embedding layer.
        """
        return embeddings.normal_embedding(num_embeddings, embedding_size)

    def predict_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        predictions, _ = self(batch)
        return predictions

    def test_step(self, batch: data.Batch, batch_idx: int) -> None:
        predictions, _ = self(batch)
        self._update_metrics(predictions, batch.target.padded)

    def training_step(self, batch: data.Batch, batch_idx: int) -> torch.Tensor:
        emissions, transitions = self(batch)
        loss = self.loss_func(batch.target.padded, emissions, transitions)
        self.log(
            "train_loss",
            loss,
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
        )
        return loss

    def validation_step(self, batch: data.Batch, batch_idx: int) -> None:
        # TODO(#317): This is not compatible with training loss.
        predictions, likelihood = self(batch)
        self.log(
            "val_loss",
            -likelihood.mean(),
            batch_size=len(batch),
            logger=True,
            on_epoch=True,
            prog_bar=True,
        )
        self._update_metrics(predictions, batch.target.padded)

    @staticmethod
    def _apply_mono_mask(transitions: torch.Tensor) -> torch.Tensor:
        """Applies monotonic attention mask to transition probabilities.

        Enforces zero log-probability values for all non-monotonic relations
        in the transition tensor (i.e., all values i < j per row j).

        Args:
            transitions (torch.Tensor): transition probabilities between
                all hidden states of shape B x src_len x src_len.

        Returns:
            torch.Tensor: masked transition probabilities of shape
                B x src_len x src_len.
        """
        mask = torch.ones_like(transitions[0]).triu().unsqueeze(0)
        # Using 0 log-probability value for masking; this is borrowed from the
        # original implementation.
        mask = (mask - 1) * defaults.NEG_LOG_EPSILON
        transitions = transitions + mask
        transitions = transitions - transitions.logsumexp(dim=2, keepdim=True)
        return transitions

    @property
    def decoder_input_size(self) -> int:
        # We concatenate along the encoding dimension.
        if self.has_features_encoder:
            return (
                self.source_encoder.output_size
                + self.features_encoder.output_size
            )
        else:
            return self.source_encoder.output_size

    @abc.abstractmethod
    def get_decoder(self) -> modules.HardAttentionRNNDecoder: ...

    @property
    @abc.abstractmethod
    def name(self) -> str: ...


class HardAttentionGRUModel(HardAttentionRNNModel):
    """Hard attention with GRU backend."""

    def get_decoder(self):
        if self.attention_context > 0:
            return modules.ContextHardAttentionGRUDecoder(
                attention_context=self.attention_context,
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )
        else:
            return modules.HardAttentionGRUDecoder(
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )

    @property
    def name(self) -> str:
        return "hard attention GRU"


class HardAttentionLSTMModel(HardAttentionRNNModel):
    """Hard attention with LSTM backend."""

    def get_decoder(self):
        if self.attention_context > 0:
            return modules.ContextHardAttentionLSTMDecoder(
                attention_context=self.attention_context,
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                hidden_size=self.decoder_hidden_size,
                embedding_size=self.embedding_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )
        else:
            return modules.HardAttentionLSTMDecoder(
                decoder_input_size=self.decoder_input_size,
                dropout=self.decoder_dropout,
                embedding_size=self.embedding_size,
                hidden_size=self.decoder_hidden_size,
                layers=self.decoder_layers,
                num_embeddings=self.target_vocab_size,
            )

    @property
    def name(self) -> str:
        return "hard attention LSTM"



"""Yoyodyne: small-vocabulary sequence-to-sequence generation."""

import warnings

import numpy
from torch import serialization

# Registers numpy objects as safe.
serialization.add_safe_globals([numpy._core.multiarray.scalar])

# Silences some stupid warnings.
warnings.filterwarnings(
    "ignore",
    ".*adds dropout after all but last recurrent layer.*",
)
warnings.filterwarnings("ignore", ".*is a wandb run already in progress.*")



"""Retrieves hyperparameters from a W&B run."""

import argparse
import logging
from typing import Any, Dict

import wandb
import yaml

from .. import util

# Top-level categories allowed in Lightning CLI configs.
ALLOWED_TOP_LEVEL_CATEGORIES = frozenset(
    ["data", "model", "trainer", "checkpoint", "prediction"]
)


def dot_to_nested_dict(flat_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Converts dot notation dictionary to nested dictionary.

    Args:
        flat_dict: Dictionary with dot-separated keys.

    Returns:
        Nested dictionary with proper structure.
    """
    nested = {}
    for key, value in flat_dict.items():
        if (
            any(
                key.startswith(f"{cat}.")
                for cat in ALLOWED_TOP_LEVEL_CATEGORIES
            )
            or key in ALLOWED_TOP_LEVEL_CATEGORIES
        ):
            util.recursive_insert(nested, key, value)
    return nested


def main() -> None:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "run_path", help="Run path (in the form entity/project/run_id)"
    )
    args = parser.parse_args()
    api = wandb.Api()
    run = api.run(f"{args.run_path}")
    logging.info("Run URL: %s", run.url)
    # Gets all config values from this run.
    flat_config = {
        key: value
        for key, value in sorted(run.config.items())
        if value is not None and "/" not in key
    }
    # Converts to nested dictionary structure.
    config = dot_to_nested_dict(flat_config)
    print(
        yaml.safe_dump(config, default_flow_style=False, sort_keys=False),
        end="",
    )



"""Encodes and decodes tensors."""

from __future__ import annotations

import dataclasses

from typing import Iterable, List

import torch

from . import indexes
from .. import special


@dataclasses.dataclass
class Mapper:
    """Handles mapping between strings and tensors."""

    index: indexes.Index  # Usually copied from the DataModule.

    @classmethod
    def read(cls, model_dir: str) -> Mapper:
        """Loads mapper from an index.

        Args:
            model_dir (str).

        Returns:
            Mapper.
        """
        return cls(indexes.Index.read(model_dir))

    # Encoding.

    def _encode(self, symbols: Iterable[str]):
        """Encodes a tensor.

        Args:
            ymbols (Iterable[str]).

        Returns:
            torch.Tensor: the encoded tensor.
        """
        return torch.tensor([self.index(symbol) for symbol in symbols])

    def encode_source(self, symbols: Iterable[str]) -> torch.Tensor:
        """Encodes a source string, padding with start and end tags.

        Args:
            symbols (Iterable[str]).

        Returns:
            torch.Tensor.
        """
        wrapped = [special.START]
        wrapped.extend(symbols)
        wrapped.append(special.END)
        return self._encode(wrapped)

    def encode_features(self, symbols: Iterable[str]) -> torch.Tensor:
        """Encodes a features string.

        Args:
            symbols (Iterable[str]).

        Returns:
            torch.Tensor.
        """
        return self._encode(symbols)

    def encode_target(self, symbols: Iterable[str]) -> torch.Tensor:
        """Encodes a features string, padding with end tags.

        Args:
            symbols (Iterable[str]).

        Returns:
            torch.Tensor.
        """
        wrapped = list(symbols)
        wrapped.append(special.END)
        return self._encode(wrapped)

    # Decoding.

    def _decode(
        self,
        indices: torch.Tensor,
    ) -> List[str]:
        """Decodes a tensor.

        Decoding halts at END; other special symbols are omitted.

        Args:
            indices (torch.Tensor): 1d tensor of indices.

        Yields:
            List[str]: Decoded symbols.
        """
        symbols = []
        for idx in indices:
            if idx == special.END_IDX:
                return symbols
            elif not special.isspecial(idx):
                symbols.append(self.index.get_symbol(idx))
        return symbols

    # These are here for compatibility; they all have the same implementation.

    def decode_source(
        self,
        indices: torch.Tensor,
    ) -> List[str]:
        return self._decode(indices)

    def decode_features(
        self,
        indices: torch.Tensor,
    ) -> List[str]:
        return self._decode(indices)

    def decode_target(
        self,
        indices: torch.Tensor,
    ) -> List[str]:
        return self._decode(indices)



"""Symbol constants and helpers."""

# These are all reserved for internal use.

UNK = "<UNK>"
UNK_IDX = 0

PAD = "<P>"
PAD_IDX = 1

START = "<S>"
START_IDX = 2

END = "<E>"
END_IDX = 3

# Keep in above order.
SPECIAL = [UNK, PAD, START, END]


def isspecial(idx: int) -> bool:
    """Determines if a symbol is in the special range.

    Args:
        idx (int):

    Returns:
        bool.
    """
    return idx < len(SPECIAL)



"""Embedding initialization functions."""

from torch import nn

from .. import special


def normal_embedding(num_embeddings: int, embedding_size: int) -> nn.Embedding:
    """Initializes the embeddings layer from a normal distribution.

    The pad embeddings are also zeroed out.

    Args:
        num_embeddings (int): number of embeddings.
        embedding_size (int): dimension of embeddings.

    Returns:
        nn.Embedding: embedding layer.
    """
    embedding_layer = nn.Embedding(num_embeddings, embedding_size)
    # Zeroes out pad embeddings.
    nn.init.constant_(embedding_layer.weight[special.PAD_IDX], 0.0)
    return embedding_layer


def xavier_embedding(num_embeddings: int, embedding_size: int) -> nn.Embedding:
    """Initializes the embeddings layer using Xavier initialization.

    The pad embeddings are also zeroed out.

    Args:
        num_embeddings (int): number of embeddings.
        embedding_size (int): dimension of embeddings.

    Returns:
        nn.Embedding: embedding layer.
    """
    embedding_layer = nn.Embedding(num_embeddings, embedding_size)
    # Xavier initialization.
    nn.init.normal_(embedding_layer.weight, mean=0, std=embedding_size**-0.5)
    # Zeros out pad embeddings.
    nn.init.constant_(embedding_layer.weight[special.PAD_IDX], 0.0)
    return embedding_layer



"""Base module class with PL integration."""

import abc

import lightning
from torch import nn

from ... import defaults


class BaseModule(abc.ABC, lightning.LightningModule):
    """Abstract base class for encoder and decoder modules.

    Unknown positional or keyword args from the superclass are ignored.

    Args:
        *args: ignored.
        dropout (float, optional): dropout probability.
        embedding_size (int, optional): the dimensionality of the embedding.
        **kwargs: ignored.
    """

    dropout: float
    dropout_layer: nn.Dropout
    embedding_size: int

    def __init__(
        self,
        *args,  # Ignored here.
        dropout: float = defaults.DROPOUT,
        embedding_size: int = defaults.EMBEDDING_SIZE,
        **kwargs,  # Ignored here.
    ):
        super().__init__()
        self.dropout = dropout
        self.dropout_layer = nn.Dropout(self.dropout)
        self.embedding_size = embedding_size

    @property
    @abc.abstractmethod
    def name(self) -> str: ...

    @property
    @abc.abstractmethod
    def output_size(self) -> int: ...



"""Transformer model classes."""

from typing import Optional

import torch
from torch import nn

from .. import data, defaults, special
from . import base, embeddings, modules


class TransformerModel(base.BaseModel):
    """Vanilla transformer model.

    If features are provided, the encodings are fused by concatenation of the
    features encoding with the source encoding on the sequence length
    dimension.

    Args:
        *args: passed to superclass.
        attention_heads (int, optional).
        decoder_max_length (int, optional): maximum length for decoder strings.
        teacher_forcing (bool, optional): should teacher (rather than student)
            forcing be used?
        **kwargs: passed to superclass.
    """

    # Model arguments.
    attention_heads: int
    teacher_forcing: bool
    classifier: nn.Linear

    def __init__(
        self,
        *args,
        attention_heads: int = defaults.ATTENTION_HEADS,
        decoder_max_length: int = defaults.MAX_LENGTH,
        teacher_forcing: bool = defaults.TEACHER_FORCING,
        **kwargs,
    ):
        self.attention_heads = attention_heads
        self.decoder_max_length = decoder_max_length
        super().__init__(*args, **kwargs)
        self.teacher_forcing = teacher_forcing
        self.classifier = nn.Linear(
            self.embedding_size, self.target_vocab_size
        )
        if (
            self.has_features_encoder
            and self.source_encoder.output_size
            != self.features_encoder.output_size
        ):
            raise base.ConfigurationError(
                "Cannot concatenate source encoding "
                f"({self.source_encoder.output_size}) and features "
                f"encoding {self.features_encoder.output_size})"
            )

    def decode_step(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        predictions: torch.Tensor,
    ) -> torch.Tensor:
        """Single decoder step.

        This predicts a distribution for one symbol.

        Args:
            source_encoded (torch.Tensor): encoded source symbols.
            source_mask (torch.Tensor): mask for the source.
            predictions (torch.Tensor): tensor of predictions thus far.

        Returns:
            torch.Tensor: logits.
        """
        # Uses a dummy mask of all zeros.
        target_mask = torch.zeros_like(predictions, dtype=bool)
        decoded, _ = self.decoder(
            encoded, mask, predictions, target_mask, self.embeddings
        )
        logits = self.classifier(decoded)
        logits = logits[:, -1, :]  # Ignores END.
        return logits

    def forward(self, batch: data.Batch) -> torch.Tensor:
        """Forward pass.

        Args:
            batch (data.Batch).

        Returns:
            torch.Tensor.

        Raises:
            base.ConfigurationError: Features encoder specified but no feature
                column specified.
            base.ConfigurationError: Features column specified but no feature
                encoder specified.
            base.ConfigurationError: Teacher forcing requested but no target
                provided.
        """
        encoded = self.source_encoder(batch.source, self.embeddings)
        mask = batch.source.mask
        if self.has_features_encoder:
            if not batch.has_features:
                raise base.ConfigurationError(
                    "Features encoder specified but "
                    "no feature column specified"
                )
            features_encoded = self.features_encoder(
                batch.features, self.embeddings
            )
            encoded = torch.cat((encoded, features_encoded), dim=1)
            mask = torch.cat((mask, batch.features.mask), dim=1)
        elif batch.has_features:
            raise base.ConfigurationError(
                "Feature column specified but no feature encoder specified"
            )
        if self.training and self.teacher_forcing:
            if not batch.has_target:
                raise base.ConfigurationError(
                    "Teacher forcing requested but no target provided"
                )
            batch_size = len(batch)
            symbol = self.start_symbol(batch_size)
            target = torch.cat((symbol, batch.target.padded), dim=1)
            target_mask = torch.cat(
                (
                    torch.ones_like(symbol, dtype=bool),
                    batch.target.mask,
                ),
                dim=1,
            )
            decoded, _ = self.decoder(
                encoded, mask, target, target_mask, self.embeddings
            )
            # -> B x target_vocab_size x seq_len.
            logits = self.classifier(decoded).transpose(1, 2)
            return logits[:, :, :-1]  # Ignores END.
        else:
            return self.greedy_decode(
                encoded,
                mask,
                batch.target.padded if batch.has_target else None,
            )

    def get_decoder(self) -> modules.TransformerDecoder:
        return modules.TransformerDecoder(
            decoder_input_size=self.source_encoder.output_size,
            dropout=self.decoder_dropout,
            embedding_size=self.embedding_size,
            hidden_size=self.decoder_hidden_size,
            layers=self.decoder_layers,
            max_length=self.decoder_max_length,
            num_embeddings=self.num_embeddings,
            attention_heads=self.attention_heads,
        )

    def greedy_decode(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        target: Optional[torch.Tensor],
    ) -> torch.Tensor:
        """Decodes the output sequence greedily.

        Args:
            encoded (torch.Tensor).
            mask (torch.Tensor).
            target (torch.Tensor, optional): target symbols; if provided
                decoding continues until this length is reached.

        Returns:
            torch.Tensor: logits from the decoder.
        """
        batch_size = mask.size(0)
        # The output distributions to be returned.
        outputs = []
        # The predicted symbols at each iteration.
        predictions = [
            torch.tensor([special.START_IDX], device=self.device).repeat(
                batch_size
            )
        ]
        if target is None:
            max_num_steps = self.max_target_length
            # Tracks when each sequence has decoded an END.
            final = torch.zeros(batch_size, device=self.device, dtype=bool)
        else:
            max_num_steps = target.size(1)
        for _ in range(max_num_steps):
            logits = self.decode_step(
                encoded,
                mask,
                torch.stack(predictions, dim=1),
            )
            outputs.append(logits)
            symbol = torch.argmax(logits, dim=1)
            predictions.append(symbol)
            if target is None:
                # Updates which sequences have decoded an END.
                final = torch.logical_or(final, (symbol == special.END_IDX))
                if final.all():
                    break
        # -> B x target_vocab_size x seq_len.
        outputs = torch.stack(outputs, dim=2)
        return outputs

    def init_embeddings(
        self, num_embeddings: int, embedding_size: int
    ) -> nn.Embedding:
        """Initializes the embedding layer.

        Args:
            num_embeddings (int): number of embeddings.
            embedding_size (int): dimension of embeddings.

        Returns:
            nn.Embedding: embedding layer.
        """
        return embeddings.xavier_embedding(num_embeddings, embedding_size)

    @property
    def max_length(self) -> int:
        if self.has_features_encoder:
            return self.max_source_length + self.max_features_length
        else:
            return self.max_source_length

    @property
    def name(self) -> str:
        return "transformer"



"""Hard attention module classes.

Hard attention models use unadulterated RNN encoders, but extend RNN decoders.
There are separate decoders for zeroth-order (HardAttentionRNNDecoder) and
first-order (ContextHardAttentionRNNDecoder) decoders."""

from typing import Tuple

import torch
from torch import nn

from ... import defaults
from . import rnn


class HardAttentionRNNDecoder(rnn.RNNDecoder):
    """Abstract base class for zeroth-order HMM hard attention RNN decoders.

    Args:
        *args: passed to superclass.
        **kwargs: passed to superclass.
    """

    output_proj: nn.Sequential
    scale_encoded: nn.Linear

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # Activates emission probs.
        self.output_proj = nn.Sequential(
            nn.Linear(self.output_size, self.output_size), nn.Tanh()
        )
        # Projects transition probabilities to depth of module.
        self.scale_encoded = nn.Linear(
            self.decoder_input_size, self.hidden_size
        )

    def forward(
        self,
        encoded: torch.Tensor,
        mask: torch.Tensor,
        symbol: torch.Tensor,
        state: rnn.RNNState,
        embeddings: nn.Embedding,
    ) -> Tuple[torch.Tensor, torch.Tensor, rnn.RNNState]:
        """Single decode pass.

        Args:
            encoded (torch.Tensor): encoded source sequence of shape
                B x seq_len x encoder_dim.
            mask (torch.Tensor): mask for the encoded source batch of
                shape B x seq_len.
            symbol (torch.Tensor): previously decoded symbol(s) of shape B x 1.
            state (RNNState): RNN state.
            embeddings (nn.Embedding): embeddings.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, RNNState]: the emission and
                transition tensors and RNN state.
        """
        embedded = self.embed(symbol, embeddings)
        decoded, state = self.module(embedded, state)
        emissions = self._get_emissions(decoded, encoded)
        transitions = self._get_transitions(decoded, encoded, mask)
        return emissions, transitions, state

    def _get_emissions(
        self, decoded: torch.Tensor, encoded: torch.Tensor
    ) -> torch.Tensor:
        """Gets emission probabilities for current timestep.

        Args:
            decoded (torch.Tensor): output from decoder for current timestep of
                shape B x 1 x decoder_dim.
            encoded (torch.Tensor): encoded source sequence of shape
                B x seq_len x encoder_dim.

        Returns:
            torch.Tensor: emissions.
        """
        output = decoded.expand(-1, encoded.size(1), -1)
        output = torch.cat((output, encoded), dim=2)
        output = self.output_proj(output)
        return output

    def _get_transitions(
        self,
        decoded: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Gets transition probabilities for current timestep.

        Given the current encoder representation and the decoder
        representation at the current time step, this calculates the alignment
        scores between all potential source sequence pairings. These
        alignments are used to predict the likelihood of state transitions
        for the output.

        After:
            Wu, S. and Cotterell, R. 2019. Exact hard monotonic attention for
            character-level transduction. In _Proceedings of the 57th Annual
            Meeting of the Association for Computational Linguistics_, pages
            1530-1537.

        Args:
            decoded (torch.Tensor): output from decoder for current timesstep
                of shape B x 1 x decoder_dim.
            encoded (torch.Tensor): encoded source sequence of shape
                B x seq_len x encoder_dim.
            mask (torch.Tensor): mask of shape B x seq_len.

        Returns:
            torch.Tensor: alignment scores across the source sequence of shape
                B x seq_len.
        """
        alignment_scores = torch.bmm(
            self.scale_encoded(encoded), decoded.transpose(1, 2)
        ).squeeze(2)
        # Gets probability of alignments.
        alignment_probs = nn.functional.softmax(alignment_scores, dim=1)
        # Masks padding.
        alignment_probs = alignment_probs * (~mask) + defaults.EPSILON
        alignment_probs = alignment_probs / alignment_probs.sum(
            dim=1, keepdim=True
        )
        # Expands over all time steps; uses log probabilities for quicker
        # computations.
        return (
            alignment_probs.log().unsqueeze(1).expand(-1, encoded.size(1), -1)
        )

    @property
    def output_size(self) -> int:
        return self.decoder_input_size + self.hidden_size


class HardAttentionGRUDecoder(HardAttentionRNNDecoder, rnn.GRUDecoder):
    """Zeroth-order HMM hard attention GRU decoder."""

    def get_module(self) -> rnn.WrappedGRUDecoder:
        return rnn.WrappedGRUDecoder(
            self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=False,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    @property
    def name(self) -> str:
        return "hard attention GRU"


class HardAttentionLSTMDecoder(HardAttentionRNNDecoder, rnn.LSTMDecoder):
    """Zeroth-order HMM hard attention LSTM decoder."""

    def get_module(self) -> rnn.WrappedLSTMDecoder:
        return rnn.WrappedLSTMDecoder(
            self.embedding_size,
            self.hidden_size,
            batch_first=True,
            bidirectional=False,
            dropout=self.dropout,
            num_layers=self.layers,
        )

    @property
    def name(self) -> str:
        return "hard attention LSTM"


class ContextHardAttentionRNNDecoder(HardAttentionRNNDecoder):
    """Abstract base class for first-order HMM hard attention RNN decoder.

    This overrides the definition of _get_transitions.
    """

    def __init__(self, attention_context: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.delta = attention_context
        # The window size must include center and both sides.
        self.alignment_proj = nn.Linear(
            self.hidden_size * 2, (self.delta * 2) + 1
        )

    def _get_transitions(
        self,
        decoded: torch.Tensor,
        encoded: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Gets transition probabilities for current timestep.

        Args:
            decoded (torch.Tensor): output from decoder for current timesstep
                of shape B x 1 x decoder_dim.
            encoded (torch.Tensor): encoded input sequence of shape
                B x seq_len x encoder_dim.
            mask (torch.Tensor): mask for the encoded input batch of
                shape B x seq_len.

        Returns:
            torch.Tensor: alignment scores across the source sequence of shape
                B x seq_len.
        """
        # Matrix multiplies encoding and decoding for alignment
        # representations. See: https://aclanthology.org/P19-1148/.
        # Expands decoded so it can concatenate with alignments.
        decoded = decoded.expand(-1, encoded.size(1), -1)
        # -> B x seq_len.
        alignment_scores = torch.cat(
            (self.scale_encoded(encoded), decoded), dim=2
        )
        alignment_scores = self.alignment_proj(alignment_scores)
        alignment_probs = nn.functional.softmax(alignment_scores, dim=1)
        # Limits context to a window of the size delta (i.e., the context
        # length).
        alignment_probs = alignment_probs.split(1, dim=1)
        alignment_probs = torch.cat(
            [
                nn.functional.pad(
                    t,
                    (
                        -self.delta + i,
                        mask.size(1) - (self.delta + 1) - i,
                    ),
                )
                for i, t in enumerate(alignment_probs)
            ],
            dim=1,
        )
        # Gets probability of alignments, masking padding.
        alignment_probs = (
            alignment_probs * (~mask).unsqueeze(1) + defaults.EPSILON
        )
        alignment_probs = alignment_probs / alignment_probs.sum(
            dim=2, keepdim=True
        )
        # Uses log probabilities for quicker computations.
        return alignment_probs.log()


class ContextHardAttentionGRUDecoder(
    ContextHardAttentionRNNDecoder, HardAttentionGRUDecoder
):
    """First-order HMM hard attention GRU decoder."""

    @property
    def name(self) -> str:
        return "contextual hard attention GRU"


class ContextHardAttentionLSTMDecoder(
    ContextHardAttentionRNNDecoder, HardAttentionLSTMDecoder
):
    """First-order HMM hard attention LSTM decoder."""

    @property
    def name(self) -> str:
        return "contextual hard attention LSTM"
